unit FileVersion;

interface
uses
  Windows, Classes;


type
  LanguageCharSet = record
                      Language: word;
                      CharSet: word;
                    end;
  PLanguageCharSet = ^LanguageCharSet;


TFileVersion = class(TPersistent)
private
  FInfoSize: longInt;
  FFileInfo: PVSFixedFileInfo;
  FError: string;
  FInfoPointer: pointer;
  FFixedFileInfo: PVSFixedFileInfo;

  FLanguagePointer: PLanguageCharSet;
  FFileName: string;
//    FFileFlags: dword;
//    FFileFlagsMask: dword;

  function GetMajorVersion: word;
  function GetMinorVersion: word;
  function GetBuildVersion: word;
  function GetReleaseVersion: word;
  procedure SetFileName(const Value: string);
  procedure FreeInfo;
    function GetLanguage: string;
    function GetCompanyName: string;
    function GetFileDescription: string;
    function GetFileVersion: string;
    function GetInternalName: string;
    function GetLegalCopyright: string;
    function GetOriginalFilename: string;
    function GetProductName: string;
    function GetProductVersion: string;
    function GetFileFlags: dword;
    function GetFileFlagsMask: dword;
public
  constructor Create;
  destructor Destroy; override;
published
  function InfoString(Item: string): string;
  property FileName: string read FFileName write SetFileName;
  property MajorVersion: word read GetMajorVersion;
  property MinorVersion: word read GetMinorVersion;
  property Release:      word read GetReleaseVersion;
  property Build:        word read GetBuildVersion;
  property FileFlagsMask: dword read GetFileFlagsMask;
  property FileFlags: dword read GetFileFlags;

  property Error: string read FError;
  property Language: string read GetLanguage;
  property CompanyName: string read GetCompanyName;
  property FileDescription: string read GetFileDescription;
  property FileVersion: string read GetFileVersion;
  property InternalName: string read GetInternalName;
  property LegalCopyright: string read GetLegalCopyright;
  property OriginalFilename: string read GetOriginalFilename;
  property ProductName: string read GetProductName;
  property ProductVersion: string read GetProductVersion;
end;

implementation
uses
  SysUtils;

{ TFileVersion }


constructor TFileVersion.Create;
begin
  inherited Create;
  FError:= 'Empty FileName'
end;

procedure TFileVersion.FreeInfo;
begin
  if (FInfoSize > 0) and (FInfoPointer <> nil) then
  begin
    FreeMem(FInfoPointer, FInfoSize);
    FInfoPointer:= nil;
    FInfoSize:= 0;
  end;
end;


destructor TFileVersion.Destroy;
begin
  FreeInfo;
  inherited Destroy
end;


procedure TFileVersion.SetFileName(const Value: string);
var
  aHandle: DWord;
  S: string;
  PISize: UInt;
begin
  FreeInfo;

  FInfoSize := GetFileVersionInfoSize(PChar(Value), aHandle);
  if FInfoSize = 0 then
  begin
    FError:= 'InfoSize=0';
    exit;
  end;

  GetMem(FInfoPointer, FInfoSize);
  if not GetFileVersionInfo(PChar(Value), 0, FInfoSize, FInfoPointer) then
  begin
    FError:= 'No FileVersionInfo';
    FreeInfo;
    exit;
  end;

  S := '\';
  if  VerQueryValue(FInfoPointer, PChar(S),Pointer(FFixedFileInfo), PISize) then
  begin
    if PISize = 0 then
    begin
      FError:= 'No FixedInfo';
      FreeInfo;
      exit;
    end
    else
    if PISize = 52 then
    begin
      FError:= '';
    end
    else
    begin
      FError:= 'Bad size of FixedInfo';
      FreeInfo;
      exit;
    end
  end;

  S:= '\VarFileInfo\Translation';
  if  VerQueryValue(FInfoPointer, PChar(S), Pointer(FLanguagePointer), PISize) then
  begin
    if PISize < 4 then
    begin
      FError:= 'No Language Info';
      FreeInfo;
      exit
    end
  end;
end;


function TFileVersion.GetMajorVersion: word;
begin
  result:= 0;
  if FError <> '' then exit;
  result:= (FFixedFileInfo.dwFileVersionMS and $FFFF0000) shr 16;
end;

function TFileVersion.GetMinorVersion: word;
begin
  result:= 0;
  if FError <> '' then exit;
  result:= (FFixedFileInfo.dwFileVersionMS and $FFFF);
end;


function TFileVersion.GetReleaseVersion: word;
begin
  result:= 0;
  if FError <> '' then exit;
  result:= (FFixedFileInfo.dwFileVersionLS and $FFFF0000) shr 16;
end;


function TFileVersion.GetBuildVersion: word;
begin
  result:= 0;
  if FError <> '' then exit;
  result:= (FFixedFileInfo.dwFileVersionLS and $FFFF);
end;



function TFileVersion.InfoString(Item: string): string;
var
  S: string;
  PPolozka: Pchar;
  PISize: UInt;
begin
  result:= '';
  if FError <> '' then exit;
  S:= '\StringFileInfo\' + IntToHex(FLanguagePointer.Language,4)
                         + IntToHex(FLanguagePointer.CharSet,4) + '\'
                         + Item;

  if  VerQueryValue(FInfoPointer, PChar(S), pointer(PPolozka), PISize) then
  begin
    if PISize = 0 then
    begin
      result:= 'No '+ Item;
    end
    else
    begin
      result:= PPolozka;
    end
  end;
end;


function TFileVersion.GetLanguage: string;
var
  aNum: longInt;
  aZnaky: array[0..50] of char;
begin
  result:= '';
  if FError <> '' then exit;
  aNum:= VerLanguageName(FLanguagePointer.Language, aZnaky, 49);
  if aNum < 50 then
     result:= aZnaky
end;

function TFileVersion.GetCompanyName: string;
begin
  result:= InfoString('CompanyName')
end;

function TFileVersion.GetFileDescription: string;
begin
  result:= InfoString('FileDescription')
end;

function TFileVersion.GetFileVersion: string;
begin
  result:= InfoString('FileVersion')
end;

function TFileVersion.GetInternalName: string;
begin
  result:= InfoString('InternalName')
end;

function TFileVersion.GetLegalCopyright: string;
begin
  result:= InfoString('LegalCopyright')
end;

function TFileVersion.GetOriginalFilename: string;
begin
  result:= InfoString('OriginalFilename')
end;

function TFileVersion.GetProductName: string;
begin
  result:= InfoString('ProductName')
end;

function TFileVersion.GetProductVersion: string;
begin
  result:= InfoString('ProductVersion')
end;

function TFileVersion.GetFileFlags: dword;
begin
  result:= 0;
  if FError <> '' then exit;
  result:= FFixedFileInfo.dwFileFlags;
end;

function TFileVersion.GetFileFlagsMask: dword;
begin
  result:= 0;
  if FError <> '' then exit;
  result:= FFixedFileInfo.dwFileFlagsMask;
end;

end.
