unit pfgDbSave;
{**************************************************************************}
{* pfgDbSave Unit                                                         *}
{*                                                                        *}
{* This unit implements the saving and loading of Palm PDB files. It also *}
{* provides a public interface method that takes in a database handle,    *}
{* and uses the database saving functionality of the object to create an  *}
{* image of the database as a file.                                       *}
{*                                                                        *}
{* Copyright (C) 2000 - 2002 by Paul Gilbert, All Rights Reserved         *}
{**************************************************************************}

interface

{$I pfgPalmConduits.inc}

uses Classes, SysUtils, pfgPalmMisc, Db, pfgPalmDb;

type
  TDbOpenMode = (dbReadOnly, dbWriteOnly, dbReadWrite);

  TpfgPalmDbFile = class
  private
    // General properties
    FFilename: string;
    FOpenMode: TDbOpenMode;
    FModified: Boolean;
    FAppInfo: TpfgModifiedMemoryStream;
    FSortInfo: TpfgModifiedMemoryStream;
    FRecordList: TList;

    // Database flags
    FDbName: string;
    FAttributes: Word;
    FVersion: Word;
    FCreationDate: TDateTime;
    FModificationDate: TDateTime;
    FLastBackupDate: TDateTime;
    FModificationNumber: LongWord;
    FDbType: LongWord;
    FCreator: LongWord;
    FUniqueIDSeed: LongWord;

    // Property access methods
    function GetModified: Boolean;
    procedure SetModified(Value: Boolean);
    function GetRecordCount: Integer;
    procedure SetRecordCount(Value: Integer);

    // Internal methods
    procedure Read;
    procedure Write;
  public
    constructor Create(AFilename: string; AOpenMode: TDbOpenMode);
    destructor Destroy; override;
    procedure Flush;
    procedure SetRecord(Index: Integer; Attributes: Byte; ID: LongWord;
      Category: Shortint; Data: Pointer; DataSize: Integer);
    procedure GetRecord(Index: Integer; out Attributes: Byte; out ID: LongWord;
      out Category: Shortint; out Data: Pointer; out DataSize: Integer);

    // Properties
    property DbName: string read FDbName write FDbName;
    property Attributes: Word read FAttributes write FAttributes;
    property Version: Word read FVersion write FVersion;
    property CreationDate: TDateTime read FCreationDate write FCreationDate;
    property ModificationDate: TDateTime read FModificationDate write
      FModificationDate;
    property LastBackupDate: TDateTime read FLastBackupDate write
      FLastBackupDate;
    property ModificationNumber: LongWord read FModificationNumber write
      FModificationNumber;
    property DbType: LongWord read FDbType write FDbType;
    property Creator: LongWord read FCreator write FCreator;
    property UniqueIDSeed: LongWord read FUniqueIDSeed write FUniqueIdSeed;
    property AppInfo: TpfgModifiedMemoryStream read FAppInfo;
    property SortInfo: TpfgModifiedMemoryStream read FSortInfo;

    property Modified: Boolean read GetModified write SetModified;
    property RecordCount: Integer read GetRecordCount write SetRecordCount;
  end;

  EPDBError = class(Exception);

  TExportTableOptions = record
    DbName: string;
    Attributes: Word;
    Version: Integer;
    CreationDate, ModificationDate, LastBackupDate: TDateTime;
    ModificationNumber: Integer;
    DbType: Integer;
    CreatorId: Integer;
    UniqueIdSeed: Integer;
    RemoteOptions: TpfgPalmRemoteTableOptions
  end;

// AssignFieldDefs
// Sets up the field definitions of the Palm table to closely match that
// of the passed DataSet

procedure AssignFieldDefs(t: TpfgPalmRemoteTable; DataSet: TDataSet);

// ExportDatabaseToFile
// These two methods use the TpfgPalmDbFile class to create images of Palm
// databases on the local computer. The first variation takes in an active
// handle to a Palm database to replicate, and the other takes in a DataSet
// whose contents should be replicated

procedure ExportDatabaseToFile(Handle: Byte; Filename: string); overload;

procedure ExportDatabaseToFile(DataSet: TDataSet; Filename: string;
  AppInfo, SortInfo: TMemoryStream; Options: TExportTableOptions);
overload;

resourcestring
  SMultipleRecordListError = 'The PDB read system cannot support PDB files ' +
    'that contain multiple linked record lists';
  SInvalidDate = 'The specified date "%s" is outside the valid date range';

implementation

uses pfgSyncMgr, pfgPalmSyncError, Dialogs;

const
  DB_NAMELEN = 32;
  eRecAttrDeleted  = $80;
  eRecAttrArchived = $08;

type
  LocalID = LongWord;

  TDatabaseHdrType = packed record
    name: Array [0..DB_NAMELEN-1] of Char;   // name of database
    attributes: Word;                        // database attributes
    version: Word;                           // version of database

    creationDate: LongWord;                  // creation date of database
    modificationDate: LongWord;              // latest modification date
    lastBackupDate: LongWord;                // latest backup date
    modificationNumber: LongWord;            // modification number of database

    appInfoID: LocalID;                      // application specific info
    sortInfoID: LocalID;                     // app specific sorting info

    dbType: LongWord;                        // database type
    creator: LongWord;                       // database creator

    uniqueIDSeed: LongWord;                  // used to generate unique IDs

    // Record list
    nextRecordListID: LocalID;               // local chunkID of next list
    numRecords: Word;                        // number of records in this list
    // first record
  end;

  TDbRecordListEntry = packed record
    DataOffset: LongWord;                    // Offset to start of data
    Attributes: Byte;                        // Attributes byte for record
    UniqueId: Array [0..2] of Byte;          // Unique Id for record
  end;
  PDbRecordListEntry = ^TDbRecordListEntry;

  TMemRecordListEntry = record
    Data: Pointer;
    DataSize: Integer;
    Attributes: Byte;
    UniqueId: LongWord;
  end;
  PMemRecordListEntry = ^TMemRecordListEntry;

{**************************************************************************}
{* Helper functions                                                       *}
{*                                                                        *}
{**************************************************************************}

const
  SecondsInDay = 60*60*24;
  SecondsInMinute = 60;
  SecondsInHour = 60*60;
  MaxRecordSize = 65000;    // Maximum size for a AppInfo/SortInfo block

// PalmLongTimeToDateTime
// Converts a longword Palm date/time (representing the number of seconds
// since midnight 1/1/1970) to a TDateTime variable. Note that the input
// should be first re-ordered for little endian

function PalmLongTimeToDateTime(Value: LongWord): TDateTime;
var
  NumDays: Integer;
  VTemp, hours, minutes, seconds: LongWord;
begin
  if Value = 0 then Result := 0
  else
  begin
    // Calculate the number of days (from 1/1/1970), and the time of day
    NumDays := Value div SecondsInDay;
    VTemp := Value mod SecondsInDay;
    hours := VTemp div SecondsInHour;
    VTemp := VTemp mod SecondsInHour;
    minutes := VTemp div SecondsInMinute;
    seconds := VTemp mod SecondsInMinute;

    // Encode final result
    Result := EncodeDate(1970, 1, 1) + NumDays +
              EncodeTime(hours, minutes, seconds, 0);
  end;
end;

// DateTimeToPalmLongTime
// Converts a Windows TDateTime variable to a longword Palm date/time value,
// which represents the number of seconds since 1/1/1970. Note that the
// output is still in little-endian format

function DateTimeToPalmLongTime(Value: TDateTime): LongWord;
var
  day, month, year: Word;
  hour, minute, second, ms: Word;
begin
  if Value = 0 then Result := 0
  else
  begin
    DecodeDate(Value, year, month, day);
    DecodeTime(Value, hour, minute, second, ms);

    // Make sure the date is in range
    if year < 1970 then raise EPDBError.CreateFmt(SInvalidDate,
      [DateTimeToStr(Value)]);

    Result := (Trunc(Value - EncodeDate(1970, 1, 1)) * SecondsInDay) +
              (hour * SecondsInHour) + (minute * SecondsInMinute) +
              second;
  end;
end;

// Reverseal functions
// These are quick mnunemics wrapping calls to the ReverseVal method

function RLW(Value: LongWord): LongWord;
begin
  Result := ReverseVal(Value, sizeof(LongWord));
end;
function RW(Value: Word): Word;
begin
  Result := ReverseVal(Value, sizeof(Word));
end;

{**************************************************************************}
{* TpfgPalmDbFile class                                                      *}
{*                                                                        *}
{**************************************************************************}

constructor TpfgPalmDbFile.Create(AFilename: string; AOpenMode: TDbOpenMode);
begin
  FRecordList := TList.Create;
  FAppInfo := TpfgModifiedMemoryStream.Create;
  FSortInfo := TpfgModifiedMemoryStream.Create;

  FFilename := AFilename;
  FOpenMode := AOpenMode;
  if AOpenMode = dbReadOnly then Read;
end;

destructor TpfgPalmDbFile.Destroy;
begin
  Flush;

  RecordCount := 0;

  FRecordList.Free;
  FAppInfo.Free;
  FSortInfo.Free;
  inherited;
end;

function TpfgPalmDbFile.GetModified: Boolean;
begin
  Result := FModified or FAppInfo.Modified or FSortInfo.Modified;
end;

procedure TpfgPalmDbFile.SetModified(Value: Boolean);
begin
  FModified := Value;
  if not Value then
  begin
    FAppInfo.Modified := False;
    FSortInfo.Modified := False;
  end;
end;

function TpfgPalmDbFile.GetRecordCount: Integer;
begin
  Result := FRecordList.Count;
end;

procedure TpfgPalmDbFile.SetRecordCount(Value: Integer);
var
  ctr: Integer;
  p: PMemRecordListEntry;
begin
  // If expanding list, then add on enough entries
  while (RecordCount < Value) do
  begin
    New(p);
    FillChar(p^, sizeof(TMemRecordListEntry), 0);
    FRecordList.Add(p);
  end;

  // If contracting the list, then reduce the number of entries
  if Value < RecordCount then
    for ctr := RecordCount-1 downto Value do
    begin
      with PMemRecordListEntry(FRecordList[ctr])^ do
        if Assigned(Data) then Dispose(Data);
      Dispose(FRecordList[ctr]);
      FRecordList.Delete(ctr);
    end;
end;

// Flush
// Wrapper for the internal Write method

procedure TpfgPalmDbFile.Flush;
begin
  if FOpenMode <> dbReadOnly then Write;
end;

// Read
// Reads the contents of a PDB file into the object

procedure TpfgPalmDbFile.Read;
var
  f: File of Byte;
  header: TDatabaseHdrType;
  TotalSize, AIOffset, AISize, SIOffset, SISize: LongWord;
  Offset: LongWord;
  RecordHeaders, PrevRecord, p: PDbRecordListEntry;
  ctr: Integer;
  PrevIndex: Integer;
  pRec: PMemRecordListEntry;
  DataSize: Integer;

  procedure ReadDataBlock(Stream: TpfgModifiedMemoryStream; Offset, Size: LongWord);
  var
    pData: PChar;
  begin
    GetMem(pData, Size);
    try
      BlockRead(f, pData^, Size);
      Stream.Clear;
      Stream.Write(pData^, Size);

    finally
      Dispose(pData);
    end;
  end;

  function ReadData(Offset, Size: LongWord): Pointer;
  var
    p: PChar;
  begin
    GetMem(p, Size);
    Seek(f, Offset);
    BlockRead(f, p^, Size);
    Result := p;
  end;

begin
  AssignFile(f, FFilename);
  Reset(f);
  try
    TotalSize := FileSize(f);

    // Parse the header data
    BlockRead(f, header, sizeof(TDatabaseHdrType));
    FDbName := StrPas(header.name);
    FAttributes := RW(header.attributes);
    FVersion := RW(header.version);
    FCreationDate := PalmLongTimeToDateTime(RLW(header.creationDate));
    FModificationDate := PalmLongTimeToDateTime(RLW(header.modificationDate));
    FLastBackupDate := PalmLongTimeToDateTime(RLW(header.lastBackupDate));
    FModificationNumber := RLW(header.modificationNumber);
    FDbType := RLW(header.dbType);
    FCreator := RLW(header.creator);
    FUniqueIdSeed := RLW(header.uniqueIDSeed);

    AIOffset := RLW(header.appInfoID);
    SIOffset := RLW(header.sortInfoID);
    AISize := 0; SISize := 0;

    if RLW(header.nextRecordListID) <> 0 then
      raise EPDBError.Create(SMultipleRecordListError);

    // Process the record list
    RecordCount := RW(header.numRecords);
    DataSize := sizeof(TDbRecordListEntry)*RecordCount + sizeof(Word);
    GetMem(RecordHeaders, DataSize);
    try
      BlockRead(f, RecordHeaders^, DataSize);
      p := RecordHeaders;

      // Firstly calculate the size of the AppInfo and SortInfo blocks. We do
      // it now, because we need to know the offset of the first record.
      Offset := RLW(p^.DataOffset);
      if AIOffset <> 0 then
      begin
        if SIOffset <> 0 then AISize := SIOffset - AIOffset
        else if Offset <> 0 then AISize := Offset - AIOffset
        else AISize := TotalSize - AIOffset;
      end;
      if SIOffset <> 0 then
      begin
        if Offset <> 0 then SISize := Offset - SIOffset
        else SISize := TotalSize - SIOffset;
      end;

      // Read in the actual data for AppInfo and SortInfo
      if AIOffset <> 0 then ReadDataBlock(FAppInfo, AIOffset, AISize);
      if SIOffset <> 0 then ReadDataBlock(FSortInfo, SIOffset, SISize);

      // Loop through the record list and read in the header and record
      // data block for each record
      PrevIndex := -1;
      PrevRecord := nil;
      for ctr := 0 to RecordCount-1 do
      begin
        with PMemRecordListEntry(FRecordList[ctr])^ do
        begin
          Attributes := p^.Attributes;
          UniqueID := (p^.UniqueID[0] shl 16) + (p^.UniqueID[1] shl 8) +
                      p^.UniqueID[2];
          Data := nil;
          DataSize := 0;

          if p^.DataOffset <> 0 then
          begin
            if Assigned(PrevRecord) then
            begin
              // Found the next record offset, so use it's offset against
              // that of the previously found record to calculate the size
              // for that previus record, and read it in
              pRec := PMemRecordListEntry(FRecordList[PrevIndex]);
              pRec^.DataSize := RLW(p^.DataOffset) - RLW(PrevRecord^.DataOffset);
              pRec^.Data := ReadData(RLW(PrevRecord^.DataOffset), pRec^.DataSize);
            end;

            PrevRecord := p;
            PrevIndex := ctr;
          end;
        end;

        Inc(p);  // Move to the next record
      end;

      // Handle any final record (which will go to the end of file)
      if PrevIndex <> -1 then
      begin
        pRec := PMemRecordListEntry(FRecordList[PrevIndex]);
        pRec^.DataSize := TotalSize - RLW(PrevRecord^.DataOffset);
        pRec^.Data :=ReadData(RLW(PrevRecord^.DataOffset), pRec^.DataSize);
      end;

    finally
      Dispose(RecordHeaders);
    end;
  finally
    CloseFile(f);
  end;
end;

// Write
// Writes out the contents of the object to a PDB file

procedure TpfgPalmDbFile.Write;
var
  f: File of Byte;
  header: TDatabaseHdrType;
  Offset: LongWord;
  RecordList, p: PDbRecordListEntry;
  ctr: Integer;
  pData: PChar;
  DataSize: Integer;

  // Writes the AppInfo or SortInfo block to file
  procedure WriteStream(Stream: TpfgModifiedMemoryStream);
  var
    pData: PChar;
  begin
    GetMem(pData, Stream.Size);
    try
      Stream.Position := 0;
      Stream.Read(pData^, Stream.Size);
      BlockWrite(f, pData^, Stream.Size);
    finally
      Dispose(pData);
    end;
  end;

begin
  AssignFile(f, FFilename);
  Rewrite(f);

  try
    // Setup the header data
    FillChar(header.name, 32, 0);
    StrPLCopy(header.name, FDbName, 31);
    header.attributes := RW(FAttributes);
    header.version := RW(FVersion);
    header.creationDate := RLW(DateTimeToPalmLongTime(FCreationDate));
    header.modificationDate := RLW(DateTimeToPalmLongTime(FModificationDate));
    header.lastBackupDate :=  RLW(DateTimeToPalmLongTime(FLastBackupDate));
    header.modificationNumber := RLW(FModificationNumber);
    header.dbType := RLW(FDbType);
    header.creator := RLW(FCreator);
    header.uniqueIDSeed := RLW(FUniqueIdSeed);
    header.nextRecordListID := 0; // No linked record list
    header.numRecords := RW(RecordCount);

    // Start of data area - accounting for the header, a single record list,
    // and the two-byte placeholder bytes at the end of the record list
    Offset := sizeof(TDatabaseHdrType) + RecordCount * sizeof(TDbRecordListEntry)
      + sizeof(Word);

    // Set offsets for AppInfo and SortInfo
    if FAppInfo.Size = 0 then header.appInfoID := 0
    else
    begin
      header.appInfoID := RLW(Offset);
      Inc(Offset, FAppInfo.Size);
    end;
    if FSortInfo.Size = 0 then header.sortInfoID := 0
    else
    begin
      header.sortInfoID := RLW(Offset);
      Inc(Offset, FSortInfo.Size);
    end;

    // Write out the header
    BlockWrite(f, header, sizeof(TDatabaseHdrType));

    // Write out the record list (if any)
    if RecordCount > 0 then
    begin
      DataSize := sizeof(TDbRecordListEntry) * RecordCount + sizeof(Word);
      GetMem(RecordList, DataSize);
      try
        FillChar(RecordList^, DataSize, 0);
        p := RecordList;
        for ctr := 0 to RecordCount-1 do
          with PMemRecordListEntry(FRecordList[ctr])^ do
          begin
            p^.Attributes := Attributes;
            p^.UniqueId[0] := UniqueId shr 16;
            p^.UniqueId[1] := (UniqueId shr 8) and $FF;
            p^.UniqueId[2] := UniqueId and $FF;

            if not Assigned(Data) then p^.DataOffset := 0
            else
            begin
              // Store the offset, and move offset to be ready for next record
              p^.DataOffset := RLW(Offset);
              Inc(Offset, DataSize);
            end;

            Inc(p);
          end;

        BlockWrite(f, RecordList^, DataSize);
      finally
        Dispose(RecordList);
      end;
    end;

    // Write out the AppInfo and SortInfo blocks if present
    if FAppInfo.Size > 0 then WriteStream(FAppInfo);
    if FSortInfo.Size > 0 then WriteStream(FSortInfo);

    // Write out each record block in turn
    for ctr := 0 to RecordCount-1 do
      with PMemRecordListEntry(FRecordList[ctr])^ do
      begin
        if Assigned(Data) then
        begin
          pData := Data;
          BlockWrite(f, pData^, DataSize);
        end;
      end;
  finally
    CloseFile(f);
  end;
end;

// SetRecord
// Sets the data for a record

procedure TpfgPalmDbFile.SetRecord(Index: Integer; Attributes: Byte;
  ID: LongWord; Category: Shortint; Data: Pointer; DataSize: Integer);
var
  p: PMemRecordListEntry;
begin
  p := PMemRecordListEntry(FRecordList[Index]);
  if Assigned(p^.Data) then Dispose(p^.Data);
  GetMem(p^.Data, DataSize);
  Move(Data^, p^.Data^, DataSize);

  p^.DataSize := DataSize;
  p^.UniqueId := ID;

  // Set up the attributes byte
  p^.Attributes := (Attributes and $F0) + (Category mod 16);
  if (Attributes and eRecAttrArchived) <> 0 then
    p^.Attributes := p^.Attributes or eRecAttrDeleted;
end;

// GetRecord
// Gets a reference to the data for a record

procedure TpfgPalmDbFile.GetRecord(Index: Integer; out Attributes: Byte;
  out ID: LongWord; out Category: Shortint; out Data: Pointer;
  out DataSize: Integer);
var
  p: PMemRecordListEntry;
begin
  p := PMemRecordListEntry(FRecordList[Index]);
  ID := p^.UniqueId;
  Data := p^.Data;
  DataSize := p^.DataSize;

  Category := p^.Attributes and $0F;
  Attributes := p^.Attributes and $F0;
  if ((Attributes and eRecAttrDeleted) <> 0) and Assigned(Data) then
    Attributes := Attributes and (not eRecAttrDeleted) or eRecAttrArchived;
end;

{**************************************************************************}
{* Support methods                                                        *}
{*                                                                        *}
{* These are suport methods for the public access methods                 *}
{**************************************************************************}

procedure LoadBlock(Handle: Byte; Stream: TpfgModifiedMemoryStream; IsAppInfo: Boolean);
var
  GenInfo: CDbGenInfo;
begin
  GetMem(GenInfo.m_pBytes, MaxRecordSize);
  try
    GenInfo.m_TotalBytes := MaxRecordSize;
    GenInfo.m_dwReserved := 0;

    if IsAppInfo then SyncReadDbAppInfoBlock(Handle, GenInfo)
    else SyncReadDbSortInfoBlock(Handle, GenInfo);

    // Store the read in data
    Stream.Clear;
    Stream.Write(GenInfo.m_pBytes^, GenInfo.m_BytesRead);
  finally
    Dispose(GenInfo.m_pBytes);
  end;
end;

{**************************************************************************}
{* Public methods                                                         *}
{*                                                                        *}
{**************************************************************************}

const
{$ifdef PFG_DELPHI4}
  FieldTypeStrs: Array [ftUnknown..ftDataSet] of string = (
    'ftUnknown', 'ftString', 'ftSmallint', 'ftInteger', 'ftWord', 'ftBoolean',
    'ftFloat', 'ftCurrency', 'ftBCD', 'ftDate', 'ftTime', 'ftDateTime',
    'ftBytes', 'ftVarBytes', 'ftAutoInc', 'ftBlob', 'ftMemo', 'ftGraphic',
    'ftFmtMemo', 'ftParadoxOle', 'ftDBaseOle', 'ftTypedBinary', 'ftCursor',
    'ftFixedChar', 'ftWideString', 'ftLargeInt', 'ftADT', 'ftArray',
    'ftReference', 'ftDataSet');
{$else}
  FieldTypeStrs: Array [ftUnknown..ftGuid] of string = (
    'ftUnknown', 'ftString', 'ftSmallint', 'ftInteger', 'ftWord', 'ftBoolean',
    'ftFloat', 'ftCurrency', 'ftBCD', 'ftDate', 'ftTime', 'ftDateTime',
    'ftBytes', 'ftVarBytes', 'ftAutoInc', 'ftBlob', 'ftMemo', 'ftGraphic',
    'ftFmtMemo', 'ftParadoxOle', 'ftDBaseOle', 'ftTypedBinary', 'ftCursor',
    'ftFixedChar', 'ftWideString', 'ftLargeInt', 'ftADT', 'ftArray',
    'ftReference', 'ftDataSet', 'ftOraBlob', 'ftOraClob', 'ftVariant',
    'ftInterface', 'ftIDispatch', 'ftGuid');
{$endif}

procedure AssignFieldDefs(t: TpfgPalmRemoteTable; DataSet: TDataSet);
var
  ctr: Integer;
begin
  t.FieldDefs.Clear;
  for ctr := 0 to DataSet.FieldDefs.Count-1 do
    with DataSet.FieldDefs[ctr] do
    begin
      case DataType of
        ftString, ftWideString: t.FieldDefs.Add(Name, ptString, Size);
        ftSmallint, ftWord: t.FieldDefs.Add(Name, ptWord);
        ftInteger, ftAutoInc: t.FieldDefs.Add(Name, ptLong);
        ftBoolean: t.FieldDefs.Add(Name, ptByte);
        ftCurrency, ftFloat: t.FieldDefs.Add(Name, ptDouble);
        ftDate: t.FieldDefs.Add(Name, ptDate);
        ftTime: t.FieldDefs.Add(Name, ptTime);
        ftDateTime: t.FieldDefs.Add(Name, ptDateTime);
        ftBytes, ftVarBytes, ftBlob: t.FieldDefs.Add(Name, ptCustom, Size);
        ftMemo, ftFmtMemo: t.FieldDefs.Add(Name, ptString, Size);
        ftLargeInt: t.FieldDefs.Add(Name, ptInt64);
      else
        raise Exception.Create(FieldTypeStrs[DataType] + ' is not supported');
      end;
    end;
end;


// ExportDatabaseToFile
// Exports the specified database file from the Palm to an image file on the
// local computer. Handle must be a handle to an open file.

procedure ExportDatabaseToFile(Handle: Byte; Filename: string);
var
  PDB: TpfgPalmDbFile;
  rParam: SyncReadOpenDbInfoParams;
  rInfo: SyncDatabaseInfoType;
  info: CRawRecordInfo;
  ctr: Integer;
begin
  // Get the database attributes
  rParam.bOptFlags := SYNC_DB_INFO_OPT_GET_ATTRIBUTES or SYNC_DB_INFO_OPT_GET_SIZE;
  rParam.bDbHandle := Handle;
  FillChar(rInfo, sizeof(SyncDatabaseInfoType), 0);
  PalmCheck(SyncReadOpenDbInfo(rParam, rInfo));

  PDB := TpfgPalmDbFile.Create(Filename, dbWriteOnly);
  try
    // Feed in the attributes
    with rInfo.baseInfo do
    begin
      PDB.DBName := StrPas(m_Name);
      PDB.Attributes := m_DbFlags;
      PDB.Version := m_Version;
      PDB.CreationDate := PalmLongTimeToDateTime(m_CreateDate);
      PDB.ModificationDate := PalmLongTimeToDateTime(m_ModDate);
      PDB.LastBackupDate := PalmLongTimeToDateTime(m_BackupDate);
      PDB.ModificationNumber := m_ModNumber;
      PDB.DbType := m_DbType;
      PDB.Creator := m_Creator;
      PDB.UniqueIDSeed := 0; // Is this correct?
    end;

    // Load in the AppInfo/SortInfo blocks
    if rInfo.dwAppBlkSize > 0 then LoadBlock(Handle, PDB.AppInfo, True);
    if rInfo.dwSortBlkSize > 0 then LoadBlock(Handle, PDB.SortInfo, False);

    // Read in the records
    GetMem(info.m_pBytes, MaxRecordSize);
    try
      PDB.RecordCount := rInfo.dwNumRecords;

      // Read in each record
      for ctr := 0 to PDB.RecordCount-1 do
      begin
        info.m_FileHandle := Handle;
        info.m_TotalBytes := MaxRecordSize;
        info.m_dwReserved := 0;
        info.m_CatId := 0;
        info.m_RecIndex := ctr;
        PalmCheck(SyncReadRecordByIndex(info));

        PDB.SetRecord(ctr, info.m_Attribs, info.m_RecId, info.m_CatId,
          info.m_pBytes, info.m_RecSize);
      end;
    finally
      Dispose(info.m_pBytes);
    end;

    // Write out the database
    PDB.Flush;

  finally
    PDB.Free;
  end;
end;

// ExportDatabaseToFile
// An alternatate version that takes in the data from a passed TDataSet
// component, and uses it to construct a Palm image file. Note that this
// method can be called outside of a HotSync session, allowing a programmer
// to pre-construct a set of tables that can be later quickly installed during
// a HotSync session

procedure ExportDatabaseToFile(DataSet: TDataSet; Filename: string;
  AppInfo, SortInfo: TMemoryStream; Options: TExportTableOptions);
var
  t: TpfgPalmRemoteTable;
  PDB: TpfgPalmDbFile;
  ctr: Integer;
  pData: Pointer;
  Size: LongWord;
  RecId: LongWord;

  procedure DuplicateStream(Src, Dest: TMemoryStream);
  var
    p: PChar;
  begin
    Dest.SetSize(Src.Size);
    Src.Position := 0; Dest.Position := 0;
    GetMem(p, Src.Size);
    try
      Src.Read(p^, Src.Size);
      Dest.Write(p^, Src.Size);
    finally
      Dispose(p);
    end;
  end;

begin
  t := TpfgPalmRemoteTable.Create(nil);
  PDB := TpfgPalmDbFile.Create(Filename, dbWriteOnly);
  try
    AssignFieldDefs(t, DataSet);
    t.DbSaving := True;

    // Set up header information
    with Options do
    begin
      PDB.DbName := DbName;
      PDB.Attributes := Attributes;
      PDB.Version := Version;
      PDB.CreationDate := CreationDate;
      PDB.ModificationDate := ModificationDate;
      PDB.LastBackupDate := LastBackupDate;
      PDB.ModificationNumber := ModificationNumber;
      PDB.DbType := DbType;
      PDB.Creator := CreatorId;
      PDB.UniqueIdSeed := UniqueIdSeed;
    end;
    t.Options := Options.RemoteOptions;
    if Assigned(AppInfo) then DuplicateStream(AppInfo, PDB.AppInfo);
    if Assigned(SortInfo) then DuplicateStream(SortInfo, PDB.SortInfo);

    RecId := 0;
    DataSet.First;
    PDB.RecordCount := DataSet.RecordCount;

    while not DataSet.Eof do
    begin
      for ctr := 0 to DataSet.FieldCount-1 do
        t.Fields[ctr].AsString := DataSet.Fields[ctr].AsString;

      t.PublicLocalToRemoteRecord(pData, Size);
      try
        // Note: currently using an incrementing counter for assigning
        // record Ids. This may need to be replaced later with something better
        Inc(RecId);
        PDB.SetRecord(DataSet.RecNo-1, 0, RecId, 0, pData, Size);
      finally
        Dispose(pData);
      end;

      DataSet.Next;
    end;

  finally
    t.Free;
    PDB.Free;
  end;
end;

end.
