unit pfgPalmComp;
{**************************************************************************}
{* pfgPalmComp Unit                                                       *}
{*                                                                        *}
{* This unit defines the TpfgPalmComponent class, a generic Palm conduit  *}
{* component that combines and replaces the old TpfgConduitRegister,      *}
{* TpfgConduitConfigure, and TpfgConduitExecute components.               *}
{*                                                                        *}
{* The component provides the following facilities:                       *}
{*   - Handles registration/unregistration of the conduit via the COM     *}
{*     DLL Server registration interface.                                 *}
{*                                                                        *}
{*   - Handles both the ConfigureConduit and CfgConduit DLL methods that  *}
{*     the Palm Sync Manager uses. If ConfigureConduit is called (ie. an  *}
{*     old version of the Sync Manager is running), then a new-style      *}
{*     properties record with as many of the fields filled in as possible *}
{*     is created, for passing to the event handler.                      *}
{*                                                                        *}
{*     By providing an event handler for the component's one event, you   *}
{*     can provide code to be executed when the "Change" button is        *}
{*     clicked in the Sync Manager list of installed conduits. The event  *}
{*     is passed the configuration structure. If you wish to determine    *}
{*     whether the structure is a full one, or a old-style constructured  *}
{*     one, check the dwVersion field: it will be $FFFFFFFF for the       *}
{*     old-style.                                                         *}
{*                                                                        *}
{*   - Handles the interfacing of conduit execution, providing an         *}
{*     OnExecConduit event that a conduit's execution code can be placed. *}
{*                                                                        *}
{* Copyright  2000-2002 by Paul Gilbert, All Rights Reserved             *}
{**************************************************************************}

interface

uses Windows, Classes, SysUtils, pfgBaseClasses, pfgCondAPI, pfgSyncMgr,
  pfgWTypes, pfgHslog;

type
  TpfgCfgConduitInfo = type CfgConduitInfoType;
  PpfgCfgConduitInfo = ^TpfgCfgConduitInfo;

  // Identical subclass definition so properties will be in unit scope when
  // the component is dropped onto a form
  TpfgSyncProperties = type CSyncProperties;

  TpfgConfigConduitProc = procedure (var Data: TpfgCfgConduitInfo) of object;

  TpfgExecConduitProc = procedure(const props: TpfgSyncProperties) of object;

  TpfgUnregisterConduitProc = procedure (Sender: TObject; var RemoveDirs:
    Boolean) of object;

  TpfgPalmComponent = class(TpfgSingletonComponent)
  private
    FOnConfigConduit: TpfgConfigConduitProc;
    FOnExecConduit: TpfgExecConduitProc;
    FOnRegisterConduit: TNotifyEvent;
    FOnUnregisterConduit: TpfgUnregisterConduitProc;

    FPriority: Integer;
    FCreatorFile: string;
    FDirectory: string;
    FConduitName: string;
    FCreatorID: string;
    FInformation: string;
    FConduitVersion: Word;
    FRemoteFile: string;

    FSyncFlag: Boolean;
    FProgressFn: PROGRESSFN;
    FProgress: string;

    function GetConduitVersion: string;
    procedure SetConduitName(const Value: string);
    procedure SetConduitVersion(const Value: string);
    procedure SetCreatorFile(const Value: string);
    procedure SetCreatorID(const Value: string);
    procedure SetDirectory(const Value: string);
    procedure SetInformation(const Value: string);
    procedure SetPriority(const Value: Integer);
    procedure DefaultConfig(var Data: TpfgCfgConduitInfo);
  protected
    procedure SetProgress(AProgress: string); virtual;
    procedure DoExecConduit(const Props: TpfgSyncProperties);
      virtual;
  published
    // Registration properties
    property Version: Word read FConduitVersion;
    property ConduitName: string read FConduitName write SetConduitName;
    property ConduitVersion: string read GetConduitVersion write SetConduitVersion;
    property Priority: Integer read FPriority write SetPriority;
    property Directory: string read FDirectory write SetDirectory;
    property CreatorFile: string read FCreatorFile write SetCreatorFile;
    property Information: string read FInformation write SetInformation;
    property RemoteFile: string read FRemoteFile write FRemoteFile;
    property CreatorID: string read FCreatorID write SetCreatorID;

    // Events
    property OnConfigConduit: TpfgConfigConduitProc read FOnConfigConduit write
      FOnConfigConduit;
    property OnExecConduit: TpfgExecConduitProc read FOnExecConduit write
      FOnExecConduit;
    property OnRegisterConduit: TNotifyEvent read FOnRegisterConduit write
      FOnRegisterConduit;
    property OnUnregisterConduit: TpfgUnregisterConduitProc
      read FOnUnregisterConduit write FOnUnregisterConduit;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    // Execution methods
    procedure Configure(var Data: TpfgCfgConduitInfo);
    procedure Log(lStr: string = ''; Act: Activity = slText;
      TimeStamp: Boolean = False);
    procedure LogFmt(lStr: string; Vals: Array of Const;
      Act: Activity = slText; TimeStamp: Boolean = False);

    // Event interface methods
    procedure DoRegisterConduit;
    function DoUnregisterConduit: Boolean;

    // Execution properties
    property Progress: string read FProgress write SetProgress;
  end;


  ERegisterError = class(Exception);

// Exported DLL methods

function DllGetClassObject(const CLSID, IID: TGUID; var Obj): HResult; stdcall;
function DllCanUnloadNow: HResult; stdcall;
function DllRegisterServer: HResult; stdcall;
function DllUnregisterServer: HResult; stdcall;

function GetConduitName(pName: PChar; nLen: Word): Integer; cdecl;
function GetConduitVersion: LongWord; cdecl;
function GetConduitInfo(infoType: ConduitInfoEnum; pIn, pOut: Pointer;
  pdwOutSize: PDWORD): Integer; cdecl;
function ConfigureConduit(var syncPrefs: CSyncPreference): Integer; cdecl;
function CfgConduit(cfgType: ConduitCfgEnum; pArgs: Pointer;
  const pdwArgsSize: PDWORD): Integer; cdecl;

function OpenConduit(progress: PROGRESSFN; const props: TpfgSyncProperties):
   Integer; cdecl; 


// Wrapper methods - these methods check for an instantiated copy of the
// Palm component, and if so, call the method of the same name within it

procedure Log(lStr: string = ''; Act: Activity = slText;
  TimeStamp: Boolean = False);
procedure LogFmt(lStr: string; Vals: Array of Const;
  Act: Activity = slText; TimeStamp: Boolean = False);

// Miscellaneous externally available methods

function DLLName: string;
function ConduitName: string;


procedure Register;

exports
  DllRegisterServer,
  DllUnregisterServer,
  DllGetClassObject,
  DllCanUnloadNow,
  GetConduitName,
  GetConduitVersion,
  GetConduitInfo,
  ConfigureConduit,
  CfgConduit,
  OpenConduit;

implementation

{$R pfgPalmIcons.res}

uses ComObj, Dialogs, pfgHotSyncAction, pfgHSAPI, pfgPalmSyncError, pfgCondMgre,
  pfgPalmDb, pfgUserData, pfgPalmMisc;

resourcestring
  SNoPalmComponent = 'The TpfgPalmComponent could not be found';

{==========================================================================}

procedure Register;
begin
  RegisterComponents('Palm Pilot', [TpfgPalmComponent]);
end;

{==========================================================================}

function PalmComponent: TpfgPalmComponent;
begin
  Result := TpfgPalmComponent.GetInstance as TpfgPalmComponent;
  if not Assigned(Result) then raise Exception.Create(SNoPalmComponent);
end;

{==========================================================================}
{= Helper methods                                                         =}
{=                                                                        =}
{= This section contains a set of miscellaneous helper methods            =}
{==========================================================================}

function DllName: string;
var
  Buffer: Array [0..255] of Char;
begin
  GetModuleFileName(HInstance, Buffer, 255);
  Result := StrPas(Buffer);
end;


function ConduitName: string;
begin
  try
    Result := PalmComponent.ConduitName;
  except
    Result := '';
  end;
end;

Procedure RemoveConduitFolders(const pCreatorID: PChar);
var
  CoreFolderSize: Integer;
  CoreFolder: Array[0..255] Of Char;
  ConduitFolderSize: Integer;
  ConduitFolder: Array[0..255] Of Char;
  NumUsers: Integer;
  UserID: LongInt;
  UserFolderSize: SmallInt;
  UserFolder: Array[0..255] Of Char;
  C: Integer;
  S: String;
begin
  CoreFolderSize:=255;
  FillChar(CoreFolder, SizeOf(CoreFolder), 0);
  if CmGetCorePath(CoreFolder, CoreFolderSize)=0 then
  try
    ConduitFolderSize:=255;
    FillChar(ConduitFolder, SizeOf(ConduitFolder), 0);
    //CmGetCreatorDirectory(PChar(obj.CreatorID), ConduitFolder, ASize): HResult;
    if CmGetCreatorDirectory(pCreatorID, ConduitFolder, ConduitFolderSize)=0 then
    try
      NumUsers:=UmGetUserCount;
      if NumUsers>0 then
        for C:=0 To NumUsers-1 do
          if UmGetUserID(C, UserID)=0 then
          try
            UserFolderSize:=255;
            FillChar(UserFolder, SizeOf(UserFolder), 0);
            If UmGetUserDirectory(UserID, UserFolder, UserFolderSize)>-1 then
            begin
              S:=StrPas(CoreFolder);
              If S[Length(S)]<>'\' then
                S:=S+'\';
              S:=S+UserFolder;
              If S[Length(S)]<>'\' then
                S:=S+'\';
              S:=S+ConduitFolder;
              RemoveDirTree(S);
            end;
          except
          end
    except
    end;
  except
  end;
end;

{==========================================================================}
{= Singleton mapped methods                                               =}
{=                                                                        =}
{= This section contains a set of miscellaneous helper methods that map   =}
{= directly onto methods of the same name on the TpfgPalmComponent class  =}
{==========================================================================}

procedure Log(lStr: string = ''; Act: Activity = slText;
  TimeStamp: Boolean = False);
begin
  try
    PalmComponent.Log(lStr, Act, TimeStamp);
  except
    raise EpfgSingletonError.Create(SSingletonNotFoundError);
  end;
end;

procedure LogFmt(lStr: string; Vals: Array of Const;
  Act: Activity = slText; TimeStamp: Boolean = False);
begin
  try
    PalmComponent.LogFmt(lStr, Vals, Act, TimeStamp);
  except
    raise EpfgSingletonError.Create(SSingletonNotFoundError);
  end;
end;


{**************************************************************************}
{* TpfgPalmComponent class                                                *}
{*                                                                        *}
{**************************************************************************}


{==========================================================================}
{= Initialization / Destruction                                           =}
{=                                                                        =}
{= This section contains the overall component management routines        =}
{==========================================================================}

constructor TpfgPalmComponent.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  // Registration properties
  FConduitName := '';
  FConduitVersion := $0100;  // 1.0
  FCreatorID := 'STRT';      // Default creator ID
  FPriority := 2;            // Default priority
  FDirectory := 'Default';   // Subdirectory to use
  FCreatorFile := '';        // Local data file
  FRemoteFile := '';         // Remote file used

  // Configure properties
  FOnConfigConduit := nil;

  // Execution properties
  FOnExecConduit := nil;
  FSyncFlag := False;
  FProgressFn := nil;
  FProgress := '';
end;

destructor TpfgPalmComponent.Destroy;
begin
  inherited Destroy;
end;

{==========================================================================}
{= Registration methods                                                   =}
{=                                                                        =}
{= This section contains the methods for conduit registsration            =}
{==========================================================================}

// DllRegisterServer
// Called by Windows to register the COM server - here we register the conduit

function DllRegisterServer: HResult;
var
  obj: TpfgPalmComponent;
  ci: PChar;
begin
  try
    obj := PalmComponent;
  except
    // Palm component was not found, so pass back a failure
    Result := S_FALSE;
    Exit;
  end;

  Result := S_OK;
  try
    obj.DoRegisterConduit;
    
    ci := PChar(obj.CreatorID);

    PalmCheck(CmInstallCreator(ci, CONDUIT_APPLICATION));
    PalmCheck(CmSetCreatorName(ci, PTChar(DLLName)));
    PalmCheck(CmSetCreatorTitle(ci, PTChar(obj.ConduitName)));
    PalmCheck(CmSetCreatorDirectory(ci, PTChar(obj.Directory)));
    if obj.CreatorFile <> '' then
      PalmCheck(CmSetCreatorFile(ci, PTChar(obj.CreatorFile)));
    If obj.RemoteFile <> '' Then
      PalmCheck(CmSetCreatorRemote(ci, PTChar(obj.RemoteFile)));
    PalmCheck(CmSetCreatorPriority(ci, obj.Priority));
    if obj.Information <> '' then
      PalmCheck(CmSetCreatorInfo(ci, PTChar(obj.Information)));

    // Try and refresh the HotSync Manager conduit list. Note that the Palm
    // documentation says NOT to use this function, so we'll wrap it in a
    // try/except block just in case the HotSync system throws a wobbly
{$IFNDEF SAFE_CODE}
    try
      HsRefreshConduitInfo;
    except
    end;
{$ENDIF}

  except
    on e:EPalmSyncError do
    begin
      ShowMessage(e.Message);        // Show the name of the Palm error
      Result := e.Error;
    end;
    on Exception do Result := E_FAIL;
  end;
end;


function DllUnregisterServer: HResult; stdcall; export;
var
  obj: TpfgPalmComponent;
  RemoveDirs: Boolean;
begin
  try
    obj := PalmComponent;

    RemoveDirs := obj.DoUnregisterConduit;
    if RemoveDirs then
      RemoveConduitFolders(PChar(obj.CreatorID));

    if CmRemoveConduitByCreatorID(PChar(obj.CreatorID)) > 0 then
    begin
      Result := S_OK;

      // Try and refresh the HotSync Manager conduit list. Note that the Palm
      // documentation says NOT to use this function, so we'll wrap it in a
      // try/except block just in case the HotSync system throws a wobbly
{$IFNDEF SAFE_CODE}
      try
        HsRefreshConduitInfo;
      except
      end;
{$ENDIF}
    end
    else
      Result := E_FAIL;
  except
    Result := E_FAIL;
  end;
end;

function DllGetClassObject(const CLSID, IID: TGUID; var Obj): HResult; stdcall; export;
begin
  Result := CLASS_E_CLASSNOTAVAILABLE;
end;

function DllCanUnloadNow: HResult; stdcall; export;
begin
  Result := S_OK;
end;


function GetConduitName(pName: PChar; nLen: Word): Integer;
var
  obj: TpfgPalmComponent;
begin
  try
    obj := PalmComponent;
    StrLCopy(pName, PChar(obj.ConduitName), nLen-1);
    Result := 0;             // Success
  except
    Result := -1;            // Failure
  end;
end;


function GetConduitVersion: LongWord;
var
  obj: TpfgPalmComponent;
begin
  try
    obj := PalmComponent;
    Result := obj.Version;
  except
    Result := 0;
  end;
end;


// GetConduitInfo
// A generalised info retrieval method

function GetConduitInfo(infoType: ConduitInfoEnum; pIn, pOut: Pointer;
  pdwOutSize: PDWORD): Integer;
type
  PeSyncTypes = ^eSyncTypes;
var
  obj: TpfgPalmComponent;
  info: PConduitRequestInfoType;
begin
  // Get a reference to the Palm component
  try
    obj := PalmComponent;
  except
    Result := S_FALSE;
    Exit;
  end;

  // Validate the parameters
  if not Assigned(pOut) then
  begin
    Result := CONDERR_INVALID_PTR;
    Exit;
  end
  else if not Assigned(pdwOutSize) then
  begin
    Result := CONDERR_INVALID_OUTSIZE_PTR;
    Exit;
  end;

  case infoType of
    eConduitName:
    begin
      if not Assigned(pIn) then
        Result := CONDERR_INVALID_INARGS_PTR
      else
      begin
        info := PConduitRequestInfoType(pIn);
        if (info^.dwVersion <> CONDUITREQUESTINFO_VERSION_1) or
           (info^.dwSize <> SZ_CONDUITREQUESTINFO) then
          Result := CONDERR_INVALID_INARGS_STRUCT
        else
        begin
          StrLCopy(PChar(pOut), PChar(obj.ConduitName), pdwOutSize^-1);
          Result := S_OK;
        end;
      end;
    end;

    eMfcVersion:
    begin
      // Since this is Delphi, the MFC is never going to be used
      if pdwOutSize^ <> sizeof(DWORD) then
        Result := CONDERR_INVALID_BUFFER_SIZE
      else
      begin
        PInteger(pOut)^ := MFC_NOT_USED;
        Result := S_OK;
      end;
    end;

    eDefaultAction:
    begin
      if pdwOutSize^ <> sizeof(eSyncTypes) then
        Result := CONDERR_INVALID_BUFFER_SIZE
      else
      begin
        PeSyncTypes(pOut)^ := eFast;  // Default start sync type
        Result := S_OK;
      end;
    end;

  else
    Result := CONDERR_UNSUPPORTED_CONDUITINFO_ENUM;
  end;
end;

function TpfgPalmComponent.GetConduitVersion: string;
begin
  Result := Format('%d.%d', [hi(FConduitVersion), lo(FConduitVersion)]); //PEL
end;

procedure TpfgPalmComponent.SetConduitName(const Value: string);
begin
  FConduitName := Value;
end;

procedure TpfgPalmComponent.SetConduitVersion(const Value: string);
var
  p: Integer;
  Major, Minor: Integer;
begin
  p := Pos('.', Value);
  if p <> 0 then
    try
      Major := StrToInt(Copy(Value, 1, p-1));
      Minor := StrToInt(Copy(Value, p+1, Length(Value)-p));

      // Validate the major/minor version numbers are correct
      if (Major >= 0) and (Major <= 255) and (Minor >= 0) and (Minor <= 255) then
      begin
        FConduitVersion := Major shl 8 + Minor;
        Exit;
      end;

    except
    end;

  // If this point reached, either via a suppressed conversion exception,
  // an invalid number, or no '.' being found, then raise a register exception
  raise ERegisterError.Create(SConversionError);
end;

procedure TpfgPalmComponent.SetCreatorFile(const Value: string);
begin
  FCreatorFile := Value;
end;

procedure TpfgPalmComponent.SetCreatorID(const Value: string);
begin
  if Length(Value) <> 4 then
    raise ERegisterError.Create(SCreatorIDError);

  FCreatorID := Value;
  pfgPalmDb.ConduitCreatorID := Value;
end;

procedure TpfgPalmComponent.SetDirectory(const Value: string);
begin
  FDirectory := Value;
end;

procedure TpfgPalmComponent.SetInformation(const Value: string);
begin
  FInformation := Value;
end;

procedure TpfgPalmComponent.SetPriority(const Value: Integer);
begin
  if (Value < 0) or (Value > 4) then
    raise ERegisterError.Create(SPriorityError);
  FPriority := Value;
end;

{==========================================================================}
{= Configuration methods                                                  =}
{=                                                                        =}
{= This section contains the configuration related methods                =}
{==========================================================================}

// Configure
// Interface method for configuration. Calls an attached OnConfigConduit
// handler if present, or otherwise defaults to the default configuration dialog

procedure TpfgPalmComponent.Configure(var Data: TpfgCfgConduitInfo);
begin
    if Assigned(OnConfigConduit) then
      OnConfigConduit(Data)
    else
      DefaultConfig(Data);
end;

// DefaultConfig
// Default configuration method which is called when none is attached to
// the ConduitConfigure's OnConfigConduit event

procedure TpfgPalmComponent.DefaultConfig(var Data: TpfgCfgConduitInfo);
begin
  with TpfgHotSyncActionDialog.Create(nil) do
    try
      case Data.syncTemporary of
        eDoNothing: ActionIndex := 3;
        eHHtoPC: ActionIndex := 2;
        ePCtoHH: ActionIndex := 1;
      else
        ActionIndex := 0;
      end;

      ConduitName:=FConduitName;
      
      // Show the dialog
      if Execute then
      begin
        case ActionIndex of
          0: Data.syncNew := eFast;
          1: Data.syncNew := ePCtoHH;
          2: Data.syncNew := eHHtoPC;
          3: Data.syncNew := eDoNothing;
        end;
      end;

      if SetAsDefault then Data.syncPref := ePermanentPreference
      else Data.syncPref := eTemporaryPreference;
    finally
      Free;
    end;
end;

function ExecConduit(var Data: TpfgCfgConduitInfo): Integer;
var
  obj: TpfgPalmComponent;
begin
  // Get a reference to the palm component
  try
    obj := PalmComponent;
    obj.Configure(data);

    Result := 0;
  except
    Result := -1;
  end;
end;


function CfgConduit(cfgType: ConduitCfgEnum; pArgs: Pointer;
  const pdwArgsSize: PDWORD): Integer;
begin
  { Perform parameter validation }
  if not Assigned(pArgs) then
    Result := CONDERR_INVALID_INARGS_PTR
  else if not Assigned(pdwArgsSize) then
    Result := CONDERR_INVALID_ARGSSIZE_PTR
  else if pdwArgsSize^ <> SZ_CFGCONDUITINFO then
    Result := CONDERR_INVALID_ARGSSIZE
  else if cfgType <> eConfig1 then
    Result := CONDERR_UNSUPPORTED_STRUCT_VERSION
  else
  begin
    // Execute the configuration
    Result := ExecConduit(PpfgCfgConduitInfo(pArgs)^);
  end;
end;

function ConfigureConduit(var syncPrefs: CSyncPreference): Integer;
var
  c: TpfgCfgConduitInfo;

  // Gets the creator ID of the conduit
  function GetCreatorID: DWORD;
  var
    obj: TpfgPalmComponent;
    s: string;
  begin
    try
      obj := PalmComponent;
      s := obj.CreatorID;
      if Length(s) <> 4 then
        Result := 0
      else
        Result := PDWORD(@s[1])^;
    except
      Result := 0;
    end;
  end;

begin
  FillChar(c, Sizeof(TpfgCfgConduitInfo), 0);
  c.dwVersion := $FFFFFFFF;
  c.dwSize := Sizeof(TpfgCfgConduitInfo);
  c.syncPref := syncPrefs.m_SyncPref;
  c.syncTemporary := syncPrefs.m_SyncType;
  c.syncPermanent := syncPrefs.m_SyncType;

  Move(syncPrefs.m_PathName, c.m_PathName, BIG_PATH);
  c.dwCreatorId := GetCreatorID;

  Result := ExecConduit(c);
  if Result <> 0 then Exit;

  // Write information back
  syncPrefs.m_SyncPref := c.syncPref;
  syncPrefs.m_SyncType := c.syncNew;
end;

{==========================================================================}
{= Configuration methods                                                  =}
{=                                                                        =}
{= This section contains the configuration related methods                =}
{==========================================================================}

// OpenConduit
// DLL method called to open the conduit

function OpenConduit(progress: PROGRESSFN; const props: TpfgSyncProperties):
  Integer; cdecl; export;
var
  obj: TpfgPalmComponent;
begin
  try
    obj := PalmComponent;
  except
    Result := -1;
    Exit;
  end;

  try
    // Store a reference to the progress function
    obj.FProgressFn := progress;

    try
      obj.DoExecConduit(props);
      Result := 0;
    except
      on e:EPalmSyncError do Result := e.Error;
      on Exception do Result := -1;
    end;
  finally
    obj.FProgressFn := nil;
  end;
end;

// SetProgress
// Sets the progress string by calling the progress function that was
// passed by HoTpfgSync. Also keeps a local copy

procedure TpfgPalmComponent.SetProgress(AProgress: string);
begin
  FProgress := AProgress;
  if Assigned(FProgressFN) then
  begin
    FProgressFN(PChar(FProgress));
    SyncYieldCycles(1);
  end;
end;

// DoExecConduit
// Interface method for the OnExecConduit event. Makes sure that an event
// is supplied. Also, if the Log method has been used to start a sync,
// makes sure a matching abort or finished log entry is added

procedure TpfgPalmComponent.DoExecConduit(const Props: TpfgSyncProperties);
begin
  if not Assigned(FOnExecConduit) then Exit;

  try
    FOnExecConduit(props);
  except
    on E:Exception do
    begin
      // Sync was started, but not properly finished, so log end
      if (E.ClassType <> EAbort) and FSyncFlag then
        Log(ConduitName, slSyncAborted);
      raise;       // Outer OpenConduit will translate exception to -1 result
    end;
  end;

  // Log the sync end if event handler didn't
  if FSyncFlag then
    Log(ConduitName, slSyncFinished);
end;

// Log
// Wrapper for the AddLogString function. Using it also allows the
// DoExecConduit interface method to ensure that a sync is properly logged
// as finished

procedure TpfgPalmComponent.Log(lStr: string = ''; Act: Activity =
  slText; TimeStamp: Boolean = False);
begin
  LogAddEntry(PChar(lStr), Act, TimeStamp);
  if Act = slSyncStarted then FSyncFlag := True
  else if Act in [slSyncAborted,slSyncFinished] then FSyncFlag := False;
end;

// LogFmt
// Similar to LogAddFormattedEntry, except it uses the Delphi style open
// array parameter to allow passing of parameters for a given format string

procedure TpfgPalmComponent.LogFmt(lStr: string; Vals: Array of Const;
  Act: Activity = slText; TimeStamp: Boolean = False);
begin
  Log(Format(lStr, Vals), Act, TimeStamp);
end;

{==========================================================================}
{= Miscellaneous methods                                                  =}
{=                                                                        =}
{==========================================================================}

// DoRegisterConduit
// Calls any attached OnRegisterConduit event

procedure TpfgPalmComponent.DoRegisterConduit;
begin
  if Assigned(FOnRegisterConduit) then FOnRegisterConduit(Self);
end;

// DoUnregisterConduit
// Calls any attached OnUnregisterConduit event. The boolean return of
// this method indicates whether to remove any folders, not whether to
// unregister or not

function TpfgPalmComponent.DoUnregisterConduit: Boolean;
begin
  Result := True;
  if Assigned(FOnUnregisterConduit) then FOnUnregisterConduit(Self, Result);
end;

end.

