unit pfgPalmDb;
{**************************************************************************}
{* pfgPalmDb Unit                                                         *}
{*                                                                        *}
{* This unit implements the database component, TpfgPalmRemoteTable.      *}
{* It provides similar properties to a standard TTable component,         *}
{* allowing records to be manipulated in tables located on a Palm.        *}
{*                                                                        *}
{* Copyright (C) 2000 - 2002 by Paul Gilbert, All Rights Reserved         *}
{**************************************************************************}

interface

{$I pfgPalmConduits.inc}

{$IFDEF PFG_DELPHI6_UP}
uses Classes, SysUtils, pfgWTypes, pfgPalmMisc, Variants;
{$ELSE}
uses Classes, SysUtils, pfgWTypes, pfgPalmMisc;
{$ENDIF}

type
  // Palm Field Types:
  //   Byte     - Single byte value
  //   Word     - 2 byte value
  //   Long     - 4 byte long value
  //   Int64    - 8 byte long value
  //   Date     - DateType value
  //   Time     - TimeType value
  //   DateTime - DateTimeType value
  //   String   - string. Can be fixed length [via Size] or null terminated
  //   Single   - 4 byte single precision 'float' value
  //   Double   - 8 byte double precision 'double' type
  //   Custom   - Custom data structure [Size specifies length]
  TpfgPalmFieldType = (ptByte, ptWord, ptDate, ptTime, ptDateTime,
    ptLong, ptString, ptInt64, ptSingle, ptDouble, ptCustom);

  // Palm table options:
  //  opShowSecret  - If set, only shows secret records
  //  opArchived    - If set, only shows archived records
  //  opModified    - If set, only shows modified recods
  // opStringsAlign - If set, specifies that a field following a string field
  //                  (unless that field is itself a string field) should
  //                  be aligned to the specified alignment boundary.
  //                  The default is to ignore alignment after strings
  // opNumAlign     - If set, specifies that 2 and 4 byte numbers should be
  //                  aligned to an even word boundary
  TpfgPalmRemoteTableOption = (opShowSecret, opArchived, opModified,
    opStringsAlign, opNumAlign);
  TpfgPalmRemoteTableOptions = set of TpfgPalmRemoteTableOption;

  // Local table options:
  //  opUseRecordID - When creating a new table, insert a "RecordID" field
  //  opUseFlags    - When creating a new table, insert a "Flags" field
  //  opUseCategory - When creating a new table, insert a "Category" field
  TpfgPalmLocalTableOption = (opUseRecordID, opUseFlags, opUseCategory);
  TpfgPalmLocalTableOptions = set of TpfgPalmLocalTableOption;

  // Remote table flags [Martin Casatti]
  TpfgPalmRemoteTableFlag = (tfAppInfoDirty,tfBackupDB,tfOkToInstallNewer,
    tfCopyPrevention,tfResetAfterInstall);
  TpfgPalmRemoteTableFlags = set of TpfgPalmRemoteTableFlag;

  TpfgPalmDataSetState = (psInactive, psBrowse, psEdit, psInsert);

  TpfgPalmReadFieldEvent = procedure (Sender: TObject; pData: Pointer;
    DataSize: Integer; FieldIndex: Integer; var DataPos: LongWord;
      out Data: string) of object;

  TpfgPalmWriteFieldEvent = procedure (Sender: TObject; FieldIndex: Integer;
    out Data: Variant) of object;

const
  CARD_NUM = 0;             // Default card used for database operations
  MaxRecordSize = 65000;    // Maximum size for a record or AppInfo block
  PalmFieldSize: Array [ptByte..ptCustom] of Integer =
    (1, 2, 2, 2, 14, 4, -1, 8, 4, 8, -1); // Field sizes

type
  TpfgPalmFieldDef = class;
  TpfgPalmFieldDefs = class;
  TpfgPalmField = class;
  TpfgPalmFields = class;
  TpfgPalmTable = class;


  TpfgPalmFieldDef = class(TCollectionItem)
  private
    FFieldName: string;
    FDataType: TpfgPalmFieldType;
    FSize: Integer;

    procedure SetFieldName(AFieldName: string);
    procedure SetDataType(ADataType: TpfgPalmFieldType);
    procedure SetSize(ASize: Integer);
  protected
    function GetDisplayName: string; override;
  public
    constructor Create(Collection: TCollection); override;
    procedure Assign(Source: TPersistent); override;
  published
    property FieldName: string read FFieldName write SetFieldName;
    property DataType: TpfgPalmFieldType read FDataType write SetDataType;
    property Size: Integer read FSize write SetSize;
  end;

  TpfgPalmFieldDefs = class(TCollection)
  private
    FOwner: TpfgPalmTable;
    FAlignment: LongWord;
    FStrictFields: Boolean;
  protected
    function GetSItem(Index: Integer): TpfgPalmFieldDef; virtual;
    procedure SetSItem(Index: Integer; Value: TpfgPalmFieldDef); virtual;
    procedure SetAlignment(AAlignment: LongWord); virtual;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(AOwner: TpfgPalmTable); reintroduce;

    function GetOwner: TPersistent; override;
    function Add: TpfgPalmFieldDef; overload;
    function Add(AFieldName: string; ADataType: TpfgPalmFieldType; Size: Integer = 0):
      TpfgPalmFieldDef; overload;
    function Insert(Index: Integer): TpfgPalmFieldDef;

    property Items[Index: Integer]: TpfgPalmFieldDef read
      GetSItem write SetSItem; default;
  published
    property Alignment: LongWord read FAlignment write SetAlignment;
    property StrictFields: Boolean read FStrictFields write FStrictFields;
  end;

  TpfgPalmField = class
  private
    FIndex: Integer;
    FOwner: TpfgPalmFields;

    function GetTable: TpfgPalmTable;
    function GetFieldName: string;
    function GetAsBoolean: Boolean;
    function GetAsDateTime: TDateTime;
    function GetAsFloat: Double;
    function GetAsInteger: Integer;
    function GetAsInt64: Int64;
    function GetAsString: string;
    function GetAsVariant: Variant;
    procedure SetAsBoolean(Value: Boolean);
    procedure SetAsDateTime(Value: TDateTime);
    procedure SetAsFloat(Value: Double);
    procedure SetAsInteger(Value: Integer);
    procedure SetAsInt64(Value: Int64);
    procedure SetAsString(Value: string);
    procedure SetAsVariant(Value: Variant);
    function GetIsNull: Boolean;
  public
    constructor Create(AOwner: TpfgPalmFields; AIndex: Integer);
    procedure Clear;

    property Owner: TpfgPalmFields read FOwner;
    property Table: TpfgPalmTable read GetTable;
    property Index: Integer read FIndex;
    property FieldName: string read GetFieldName;

    property AsBoolean: Boolean read GetAsBoolean write SetAsBoolean;
    property AsDateTime: TDateTime read GetAsDateTime write SetAsDateTime;
    property AsFloat: Double read GetAsFloat write SetAsFloat;
    property AsInteger: Integer read GetAsInteger write SetAsInteger;
    property AsInt64: Int64 read GetAsInt64 write SetAsInt64;
    property AsString: string read GetAsString write SetAsString;
    property AsVariant: Variant read GetAsVariant write SetAsVariant;
    property IsNull: Boolean read GetIsNull;
  end;

  TpfgPalmFields = class(TPersistent)
  private
    FFields: Array of TpfgPalmField;
    FOwner: TpfgPalmTable;

    function GetCount: Integer;
    function GetItem(Index: Integer): TpfgPalmField;
    procedure SetItem(Index: Integer; AField: TpfgPalmField);
    function GetFlags: Byte;
    procedure SetFlags(AFlags: Byte);
    function GetRecordID: LongWord;
    procedure SetRecordID(AID: LongWord);
    function GetCategory: Shortint;
    procedure SetCategory(ACategory: Shortint);
    function GetIsEmpty: Boolean;
  public
    constructor Create(AOwner: TpfgPalmTable);
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
    procedure Clear;
    procedure Update;

    property Owner: TpfgPalmTable read FOwner;
    property Count: Integer read GetCount;
    property Items[Index: Integer]: TpfgPalmField read GetItem write SetItem; default;
    property Flags: Byte read GetFlags write SetFlags;
    property RecordID: LongWord read GetRecordID write SetRecordID;
    property Category: Shortint read GetCategory write SetCategory;
    property IsEmpty: Boolean read GetIsEmpty;
  end;


  TpfgPalmTable = class(TComponent)
  private
    FFieldDefs: TpfgPalmFieldDefs;
    FFields: TpfgPalmFields;
    FReadOnly: Boolean;

    // Event handler method pointers
    FOnBeforeClose, FOnAfterClose: TNotifyEvent;
    FOnBeforeOpen, FOnAfterOpen: TNotifyEvent;
    FOnBeforeRecChange, FOnAfterRecChange: TNotifyEvent;
    FOnBeforeEdit, FOnAfterEdit: TNotifyEvent;
    FOnBeforeInsert, FOnAfterInsert: TNotifyEvent;
    FOnBeforePost, FOnAfterPost: TNotifyEvent;
    FOnBeforeCancel, FOnAfterCancel: TNotifyEvent;
    FOnBeforeDelete, FOnAfterDelete: TNotifyEvent;
    FOnBeforeTableCreate, FOnAfterTableCreate: TNotifyEvent;
    FOnReadField: TpfgPalmReadFieldEvent;
    FOnWriteField: TpfgPalmWriteFieldEvent;

    procedure SetFieldDefs(AFieldDefs: TpfgPalmFieldDefs);
    procedure SetFields(AFields: TpfgPalmFields);
    function GetFieldCount: Integer;
  protected
    { Interfacing methods for sub-objects }
    procedure FieldDefChanged; virtual;
    function GetFieldValue(Index: Integer): Variant; virtual; abstract;
    procedure SetFieldValue(Index: Integer; Value: Variant); virtual; abstract;
    function GetRecordFlags: Byte; virtual; abstract;
    procedure SetRecordFlags(AFlags: Byte); virtual; abstract;
    function GetRecordID: LongWord; virtual; abstract;
    procedure SetRecordID(Value: LongWord); virtual; abstract;
    function GetRecordCategory: Shortint; virtual; abstract;
    procedure SetRecordCategory(ACategory: Shortint); virtual; abstract;
    function GetIsEmpty: Boolean; virtual; abstract;

    // Event handler execution methods
    procedure DoBeforeClose; dynamic;
    procedure DoAfterClose; dynamic;
    procedure DoBeforeOpen; dynamic;
    procedure DoAfterOpen; dynamic;
    procedure DoBeforeRecChange; dynamic;
    procedure DoAfterRecChange; dynamic;
    procedure DoBeforeEdit; dynamic;
    procedure DoAfterEdit; dynamic;
    procedure DoBeforeInsert; dynamic;
    procedure DoAfterInsert; dynamic;
    procedure DoBeforePost; dynamic;
    procedure DoAfterPost; dynamic;
    procedure DoBeforeCancel; dynamic;
    procedure DoAfterCancel; dynamic;
    procedure DoBeforeDelete; dynamic;
    procedure DoAfterDelete; dynamic;
    procedure DoBeforeTableCreate; dynamic;
    procedure DoAfterTableCreate; dynamic;

    procedure DoReadField(pData: Pointer; DataSize: Integer;
      FieldIndex: Integer; var DataPos: LongWord; out Data: string); dynamic;
    procedure DoWriteField(FieldIndex: Integer; out Data: Variant); dynamic;

    // Property handler methods
    function GetActive: Boolean; virtual; abstract;
    procedure SetActive(AActive: Boolean); virtual; abstract;
    function GetTableName: string; virtual; abstract;
    procedure SetTableName(AName: string); virtual; abstract;
    function GetReadOnly: Boolean; virtual;
    procedure SetReadOnly(AReadOnly: Boolean); virtual;
    function GetState: TpfgPalmDataSetState; virtual; abstract;
    function GetRecordCount: Integer; virtual; abstract;
    function GetRecNum: Integer; virtual; abstract;
    procedure SetRecNum(ARecNum: Integer); virtual; abstract;
    function GetBOF: Boolean; virtual; abstract;
    function GetEOF: Boolean; virtual; abstract;
  published
    // Properties
    property FieldDefs: TpfgPalmFieldDefs read FFieldDefs write SetFieldDefs;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly;
    property TableName: string read GetTableName write SetTableName;

    // Events
    property OnBeforeClose: TNotifyEvent read FOnBeforeClose write FOnBeforeClose;
    property OnAfterClose: TNotifyEvent read FOnAfterClose write FOnAfterClose;
    property OnBeforeOpen: TNotifyEvent read FOnBeforeOpen write FOnBeforeOpen;
    property OnAfterOpen: TNotifyEvent read FOnAfterOpen write FOnAfterOpen;
    property OnBeforeRecChange: TNotifyEvent read FOnBeforeRecChange
      write FOnBeforeRecChange;
    property OnAfterRecChange: TNotifyEvent read FOnAfterRecChange
      write FOnAfterRecChange;
    property OnBeforeEdit: TNotifyEvent read FOnBeforeEdit write FOnBeforeEdit;
     property OnAfterEdit: TNotifyEvent read FOnAfterEdit write FOnAfterEdit;
    property OnBeforeInsert: TNotifyEvent read FOnBeforeInsert write FOnBeforeInsert;
    property OnAfterInsert: TNotifyEvent read FOnAfterInsert write FOnAfterInsert;
    property OnBeforePost: TNotifyEvent read FOnBeforePost write FOnBeforePost;
    property OnAfterPost: TNotifyEvent read FOnAfterPost write FOnAfterPost;
    property OnBeforeCancel: TNotifyEvent read FOnBeforeCancel write FOnBeforeCancel;
    property OnAfterCancel: TNotifyEvent read FOnAftercancel write FOnAfterCancel;
    property OnBeforeTableCreate: TNotifyEvent read FOnBeforeTableCreate write FOnBeforeTableCreate;
    property OnAfterTableCreate: TNotifyEvent read FOnAfterTableCreate write FOnAfterTableCreate;

    property OnReadField: TpfgPalmReadFieldEvent read FOnReadField write FOnReadField;
    property OnWriteField: TpfgPalmWriteFieldEvent read FOnWriteField write FonWriteField;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Open; virtual;
    procedure Close; virtual;
    function FindField(FieldName: string): TpfgPalmField; virtual;
    function FieldByName(FieldName: string): TpfgPalmField; virtual;
    function Exists: Boolean; virtual;

    procedure EmptyTable; virtual; abstract;
    procedure CreateTable; virtual; abstract;
    procedure DeleteTable; virtual; abstract;
    procedure First; virtual; abstract;
    procedure Last; virtual; abstract;
    procedure Next; virtual; abstract;
    procedure Prior; virtual; abstract;
    procedure Edit; virtual; abstract;
    procedure Append; virtual; abstract;
    procedure Insert; virtual; abstract;
    procedure Post; virtual; abstract;
    procedure Cancel; virtual; abstract;
    procedure Delete; virtual; abstract;

    // Public runtime properties
    property Fields: TpfgPalmFields read FFields write SetFields;
    property FieldCount: Integer read GetFieldCount;
    property Active: Boolean read GetActive write SetActive;
    property State: TpfgPalmDataSetState read GetState;
    property RecordCount: Integer read GetRecordCount;
    property BOF: Boolean read GetBOF;
    property EOF: Boolean read GetEOF;
  end;

  TpfgPalmRemoteTable = class(TpfgPalmTable)
  private
    FCreatorID: string;
    FTableName: string;
    FTableType: string;
    FVersion: Word;
    FHandle: Byte;
    FOptions: TpfgPalmRemoteTableOptions;
    FAppInfo: TpfgModifiedMemoryStream;
    FCategoryFilter: Integer;
    FState: TpfgPalmDataSetState;
//    FRecord: Pointer;
//    FRecordSize: Integer;
    FRecNum: Integer;
    FBOF, FEOF: Boolean;
    FdbFlags: TpfgPalmRemoteTableFlags;
    FRecordFlags: Byte;
    FRecordID: LongWord;
    FRecordCategory: Shortint;
    FData: TStringList;
    FRecordEmpty: Boolean;
    FCardNum: Word;
    FRecordSaved: Boolean;
    FDbSaving: Boolean; // Flag used during exporting datasets to files

    // Internal methods
    procedure OpenDatabase;
    procedure CloseDatabase;
    procedure GetNextRecord;
    procedure PostRecord;
    procedure DeleteRecord(ARecID: LongWord);
    procedure InitRecord;
    procedure RemoteRecordToLocal(const pData: Pointer; Size: LongWord);
    procedure LocalToRemoteRecord(out pData: Pointer; out Size: LongWord);
    function GetDbFlagsField: Byte;
  protected
    // Embedded object interface methods
    procedure FieldDefChanged; override;
    function GetFieldValue(Index: Integer): Variant; override;
    procedure SetFieldValue(Index: Integer; Value: Variant); override;
    function GetRecordFlags: Byte; override;
    procedure SetRecordFlags(AFlags: Byte); override;
    function GetRecordID: LongWord; override;
    procedure SetRecordID(Value: LongWord); override;
    function GetRecordCategory: Shortint; override;
    procedure SetRecordCategory(ACategory: Shortint); override;
    function GetIsEmpty: Boolean; override;

    // Property handler methods
    procedure SetHandle(AHandle: Byte);
    function GetActive: Boolean; override;
    procedure SetActive(AActive: Boolean); override;
    function GetTableName: string; override;
    procedure SetTableName(AName: string); override;
    procedure SetCreatorID(AID: string); virtual;
    procedure SetTableType(AType: string); virtual;
    procedure SetVersion(AVersion: Word); virtual;
    function GetOptions: TpfgPalmRemoteTableOptions; virtual;
    procedure SetOptions(AOptions: TpfgPalmRemoteTableOptions); virtual;
    function GetAppInfo: TpfgModifiedMemoryStream; virtual;
    procedure SetAppInfo(AInfo: TpfgModifiedMemoryStream); virtual;
    function GetCategoryFilter: Shortint; virtual;
    procedure SetCategoryFilter(ACategory: Shortint); virtual;
    function GetState: TpfgPalmDataSetState; override;
    function GetRecordCount: Integer; override;
    function GetRecNum: Integer; override;
    procedure SetRecNum(ARecNum: Integer); override;
    function GetBOF: Boolean; override;
    function GetEOF: Boolean; override;
    procedure SetDbFlags(Value: TpfgPalmRemoteTableFlags); virtual;
    procedure SetDbSaving(Value: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;

    procedure EmptyTable; override;
    procedure CreateTable; override;
    procedure DeleteTable; override;
    procedure First; override;
    procedure Last; override;
    procedure Next; override;
    procedure Prior; override;
    procedure Edit; override;
    procedure Append; override;
    procedure Insert; override;
    procedure Post; override;
    procedure Cancel; override;
    procedure Delete; override;
    function Exists: Boolean; override;
    procedure DeleteById(ARecID: LongWord); virtual;
    function LocateById(ARecID: LongWord): Boolean; virtual;
    procedure PurgeTable; virtual;
    procedure SaveToFile(Filename: string);

    procedure PublicLocalToRemoteRecord(out pData: Pointer; out Size: LongWord);

    // Public run-time properties
    property Handle: Byte read FHandle write SetHandle stored False;
    property AppInfo: TpfgModifiedMemoryStream read GetAppInfo write SetAppInfo stored False;
//    property DbRecord: Pointer read FRecord;
    property RecNum: Integer read GetRecNum write SetRecNum;
    property dbFlags: TpfgPalmRemoteTableFlags read FdbFlags write SetDbFlags;
  published
    property CreatorID: string read FCreatorID write SetCreatorID;
    property TableType: string read FTableType write SetTableType;
    property Version: Word read FVersion write SetVersion;
    property Options: TpfgPalmRemoteTableOptions read GetOptions write SetOptions;
    property CategoryFilter: Shortint read GetCategoryFilter write SetCategoryFilter;
    property CardNum: Word read FCardNum write FCardNum;
    property DbSaving: Boolean read FDbSaving write SetDbSaving;
  end;


  TpfgPalmBatchMove = class(TComponent)
  private
    FSource: TpfgPalmTable;
    FDestination: TpfgPalmTable;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;

    procedure Execute;
  published
    property Source: TpfgPalmTable read FSource write FSource;
    property Destination: TpfgPalmTable read FDestination write FDestination;
  end;


  { Exceptions }

  EPalmDBError = class(Exception);

  ENotImplError = class(Exception);

  EFeatureNotAvailable = class(Exception);

var
  // This variable is automatically set by the TpfgPalmComponent when it is
  // used
  ConduitCreatorID: string = '';

procedure Register;

resourcestring
  // Shared resource strings (those which are referenced by the
  // pfgPalmLocalDb.pas file)
  SActiveError = 'The operation requires the table to be closed';

implementation

uses Windows, pfgSyncMgr, pfgPalmSyncError, pfgDbSave;

resourcestring
  // Internal resource strings
  SFieldDefsError = 'Created FieldDefs did not have a table owner';
  STableTypeError = 'The TableType "%s" is invalid';
  SNotActiveError = 'The operation requires the table to be open';
  SReadOnlyError = 'The database is read only';
  SRuntimeError = 'The operation is only allowed at runtime';
  SEditModeError = 'Cannot perform this operation whilst in edit mode';
  SCategoryError = 'The category number "%d" is invalid';
  SInvalidRecNumError = 'The Record Number "%d" could not be found';
  SFieldCountMismatchError = 'The number of fields in the two field objects ' +
    'did not match';
//  SNoActiveRecordError = 'There is no currently active record';
  SFieldMappingError = 'A mapped field went beyond the end of the physical ' +
    'record structure';
  SRecNotFound = 'The operation requires an active record';
  SFieldNotFound = 'The specified field "%s" was not found';
  SSrcDestMissingError = 'Source or Destination missing in BatchMove';
  //SRecordIDRemoteError = 'RecordID cannot be set for a remote table';

procedure Register;
begin
  RegisterComponents('Palm Pilot', [TpfgPalmRemoteTable, TpfgPalmBatchMove]);
end;


{**************************************************************************}
{* TpfgPalmFieldDefs class                                                   *}
{*                                                                        *}
{**************************************************************************}

constructor TpfgPalmFieldDefs.Create(AOwner: TpfgPalmTable);
begin
  inherited Create(TpfgPalmFieldDef);
  FOwner := AOwner;
  FAlignment := 0;  // Default to no alignment
  FStrictFields := True; // Palm records must map to field definitions exactly 
  
  if not Assigned(AOwner) then raise EPalmDBError.Create(SFieldDefsError);
end;

function TpfgPalmFieldDefs.GetSItem(Index: Integer): TpfgPalmFieldDef;
begin
  Result := inherited Items[Index] as TpfgPalmFieldDef;
end;

procedure TpfgPalmFieldDefs.SetSItem(Index: Integer; Value: TpfgPalmFieldDef);
begin
  inherited Items[Index] := Value;
end;

procedure TpfgPalmFieldDefs.SetAlignment(AAlignment: LongWord);
begin
  if AAlignment <> FAlignment then
  begin
    FAlignment := AAlignment;
    Update(nil);
  end;
end;

function TpfgPalmFieldDefs.GetOwner: TPersistent;
begin
  Result := FOwner;
end;

function TpfgPalmFieldDefs.Add: TpfgPalmFieldDef;
begin
  Result := inherited Add as TpfgPalmFieldDef;
end;

function TpfgPalmFieldDefs.Add(AFieldName: string; ADataType: TpfgPalmFieldType;
  Size: Integer = 0): TpfgPalmFieldDef;
begin
  Result := Self.Add;
  Result.FieldName := AFieldName;
  Result.DataType := ADataType;
  Result.Size := Size;
end;

function TpfgPalmFieldDefs.Insert(Index: Integer): TpfgPalmFieldDef;
begin
  Result := inherited Insert(Index) as TpfgPalmFieldDef;
end;

procedure TpfgPalmFieldDefs.Update(Item: TCollectionItem);
begin
  FOwner.FieldDefChanged;
end;

{**************************************************************************}
{* TpfgPalmFieldDef class                                                    *}
{*                                                                        *}
{**************************************************************************}

constructor TpfgPalmFieldDef.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FDataType := ptWord;
  FSize := 0;
  FFieldName := '';
end;

procedure TpfgPalmFieldDef.Assign(Source: TPersistent);
var
  s: TpfgPalmFieldDef;
begin
  if Source is TpfgPalmFieldDef then
  begin
    s := TpfgPalmFieldDef(Source);
    FDataType := s.FDataType;
    FSize := s.FSize;
    FFieldName := s.FFieldName;
  end
  else
    inherited Assign(Source);
end;

procedure TpfgPalmFieldDef.SetFieldName(AFieldName: string);
begin
  if FFieldName <> AFieldName then
  begin
    FFieldName := AFieldName;
    Changed(False);
  end;
end;

procedure TpfgPalmFieldDef.SetDataType(ADataType: TpfgPalmFieldType);
begin
  if FDataType <> ADataType then
  begin
    FDataType := ADataType;
    Changed(False);
  end;
end;

procedure TpfgPalmFieldDef.SetSize(ASize: Integer);
begin
  if FSize <> ASize then
  begin
    FSize := ASize;
    Changed(False);
  end;
end;

function TpfgPalmFieldDef.GetDisplayName: string;
begin
  if FieldName = '' then Result := inherited GetDisplayName
  else Result := FieldName;
end;

{**************************************************************************}
{* TpfgPalmField class                                                       *}
{*                                                                        *}
{**************************************************************************}

constructor TpfgPalmField.Create(AOwner: TpfgPalmFields; AIndex: Integer);
begin
  inherited Create;
  FOwner := AOwner;
  FIndex := AIndex;
end;

function TpfgPalmField.GetTable: TpfgPalmTable;
begin
  Result := FOwner.Owner;
end;

function TpfgPalmField.GetFieldName: string;
begin
  // Return the field name of the field def at the same index
  Result := Table.FieldDefs[Index].FieldName;
end;

procedure TpfgPalmField.Clear;
begin
  SetAsString('');
end;

function TpfgPalmField.GetAsBoolean: Boolean;
begin
  Result := Self.AsInteger <> 0;
end;

function TpfgPalmField.GetAsDateTime: TDateTime;
begin
  Result := StrToDateTime(Self.AsString);
end;

function TpfgPalmField.GetAsFloat: Double;
begin
  Result := Table.GetFieldValue(FIndex);
end;

function TpfgPalmField.GetAsInteger: Integer;
var
  s: string;
begin
  s := Table.GetFieldValue(FIndex);
  if s = '' then Result := 0
  else Result := StrToInt(s);
end;

function TpfgPalmField.GetAsInt64: Int64;
var
  s: string;
begin
  // Since a variant can't hold an Int64, the value will be
  // returned as a string
  s := Table.GetFieldValue(FIndex);
  if s = '' then Result := 0
  else Result := StrToInt64(s);
end;

function TpfgPalmField.GetAsString: string;
begin
  Result := Table.GetFieldValue(FIndex);
end;

function TpfgPalmField.GetAsVariant: Variant;
begin
  Result := Table.GetFieldValue(FIndex);
end;

procedure TpfgPalmField.SetAsBoolean(Value: Boolean);
begin
  if Value then Table.SetFieldValue(FIndex, 1)
  else Table.SetFieldValue(FIndex, 0);
end;

procedure TpfgPalmField.SetAsDateTime(Value: TDateTime);
begin
  Table.SetFieldValue(FIndex, DateTimeToStr(Value));
end;

procedure TpfgPalmField.SetAsFloat(Value: Double);
begin
  Table.SetFieldValue(FIndex, Value);
end;

procedure TpfgPalmField.SetAsInteger(Value: Integer);
begin
  Table.SetFieldValue(FIndex, Value);
end;

procedure TpfgPalmField.SetAsInt64(Value: Int64);
begin
  // Pass as a string, since a variant can't hold an Int64
  Table.SetFieldValue(FIndex, IntToStr(Value));
end;

procedure TpfgPalmField.SetAsString(Value: string);
begin
  Table.SetFieldValue(FIndex, Value);
end;

procedure TpfgPalmField.SetAsVariant(Value: Variant);
begin
  Table.SetFieldValue(FIndex, Value);
end;

function TpfgPalmField.GetIsNull: Boolean;
begin
  Result := Self.AsString = '';
end;

{**************************************************************************}
{* TpfgPalmFields class                                                      *}
{*                                                                        *}
{**************************************************************************}

constructor TpfgPalmFields.Create(AOwner: TpfgPalmTable);
begin
  inherited Create;
  FOwner := AOwner;
  SetLength(FFields, 0);
end;

destructor TpfgPalmFields.Destroy;
begin
  Clear;
  inherited Destroy;
end;

// Assign
// Takes care of assigning the data from one set of fields to another. The
// number of fields must match

procedure TpfgPalmFields.Assign(Source: TPersistent);
var
  s: TpfgPalmFields;
  ctr: Integer;
begin
  if not (Source is TpfgPalmFields) then
    inherited Assign(Source)
  else
  begin
    s := TpfgPalmFields(Source);
    if s.Count <> Count then raise EPalmDbError.Create(SFieldCountMismatchError);

    for ctr := 0 to Count-1 do
      Items[ctr].AsVariant := s[ctr].AsVariant;

    try
      Flags := s.Flags;
    except
    end;
    try
      //RecordID := s.RecordID; unassigned for now
    except
    end;
    try
      Category := s.Category;
    except
    end;
  end;
end;

// Clear
// Clears the current field list

procedure TpfgPalmFields.Clear;
var
  ctr: Integer;
begin
  for ctr := 0 to Count-1 do
    FFields[ctr].Free;
  SetLength(FFields, 0);
end;

// Update
// Creates a field list based on the current field definition

procedure TpfgPalmFields.Update;
var
  ctr: Integer;
begin
  Clear;
  SetLength(FFields, Owner.FieldCount);
  for ctr := 0 to High(FFields) do
    FFields[ctr] := TpfgPalmField.Create(Self, ctr);
end;

// GetCount
// Returns the number of fields

function TpfgPalmFields.GetCount: Integer;
begin
  Result := Length(FFields);
end;

// GetItem
// Returns the item at the specified index

function TpfgPalmFields.GetItem(Index: Integer): TpfgPalmField;
begin
  Result := FFields[Index];
end;

// SetItem
// Copys the value of one field into another

procedure TpfgPalmFields.SetItem(Index: Integer; AField: TpfgPalmField);
begin
  FFields[Index].AsVariant := AField.AsVariant;
end;

// GetFlags
// Returns the flags for the record

function TpfgPalmFields.GetFlags: Byte;
begin
  Result := Owner.GetRecordFlags;
end;

// SetFlags
// Sets the flags for the record

procedure TpfgPalmFields.SetFlags(AFlags: Byte);
begin
  Owner.SetRecordFlags(AFlags);
end;

function TpfgPalmFields.GetRecordID: LongWord;
begin
  Result := Owner.GetRecordID;
end;

procedure TpfgPalmFields.SetRecordID(AID: LongWord);
begin
  Owner.SetRecordID(AID);
end;

function TpfgPalmFields.GetCategory: Shortint;
begin
  Result := Owner.GetRecordCategory;
end;

procedure TpfgPalmFields.SetCategory(ACategory: Shortint);
begin
  Owner.SetRecordCategory(ACategory);
end;

function TpfgPalmFields.GetIsEmpty: Boolean;
begin
  Result := Owner.GetIsEmpty;
end;

{**************************************************************************}
{* TpfgPalmTable class                                                       *}
{*                                                                        *}
{**************************************************************************}

constructor TpfgPalmTable.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFieldDefs := TpfgPalmFieldDefs.Create(Self);
  FFields := TpfgPalmFields.Create(Self);

  // Initialize event handler pointers
  FOnBeforeClose     := nil;
  FOnAfterClose      := nil;
  FOnBeforeOpen      := nil;
  FOnAfterOpen       := nil;
  FOnBeforeRecChange := nil;
  FOnAfterRecChange  := nil;
  FOnBeforeEdit      := nil;
  FOnAfterEdit       := nil;
  FOnBeforeInsert    := nil;
  FOnAfterInsert     := nil;
  FOnBeforePost      := nil;
  FOnAfterPost       := nil;
  FOnBeforeCancel    := nil;
  FOnAfterCancel     := nil;
  FOnReadField       := nil;
  FOnWriteField      := nil;
end;

destructor TpfgPalmTable.Destroy;
begin
  FFieldDefs.Free;
  FFields.Free;
  inherited Destroy;
end;

{==========================================================================}
{= Junction methods for embedded methods                                  =}
{=                                                                        =}
{==========================================================================}

procedure TpfgPalmTable.FieldDefChanged;
begin
  // Do nothing
end;

{==========================================================================}
{= Event dispatch methods                                                 =}
{=                                                                        =}
{==========================================================================}

procedure TpfgPalmTable.DoBeforeClose;
begin
  if Assigned(FOnBeforeClose) then FOnBeforeClose(Self);
end;

procedure TpfgPalmTable.DoAfterClose;
begin
  if Assigned(FOnAfterClose) then FOnAfterClose(Self);
end;

procedure TpfgPalmTable.DoBeforeOpen;
begin
  if Assigned(FOnBeforeOpen) then FOnBeforeOpen(Self);
end;

procedure TpfgPalmTable.DoAfterOpen;
begin
  if Assigned(FOnAfterOpen) then FOnAfterOpen(Self);
end;

procedure TpfgPalmTable.DoBeforeRecChange;
begin
  if Assigned(FOnBeforeRecChange) then FOnBeforeRecChange(Self);
end;

procedure TpfgPalmTable.DoAfterRecChange;
begin
  if Assigned(FOnAfterRecChange) then FOnAfterRecChange(Self);
end;

procedure TpfgPalmTable.DoBeforeEdit;
begin
  if Assigned(FOnBeforeEdit) then FOnBeforeEdit(Self);
end;

procedure TpfgPalmTable.DoAfterEdit;
begin
  if Assigned(FOnAfterEdit) then FOnAfterEdit(Self);
end;

procedure TpfgPalmTable.DoBeforeInsert;
begin
  if Assigned(FOnBeforeInsert) then FOnBeforeInsert(Self);
end;

procedure TpfgPalmTable.DoAfterInsert;
begin
  if Assigned(FOnAfterInsert) then FOnAfterInsert(Self);
end;

procedure TpfgPalmTable.DoBeforePost;
begin
  if Assigned(FOnBeforePost) then FOnBeforePost(Self);
end;

procedure TpfgPalmTable.DoAfterPost;
begin
  if Assigned(FOnAfterPost) then FOnAfterPost(Self);
end;

procedure TpfgPalmTable.DoBeforeCancel;
begin
  if Assigned(FOnBeforeCancel) then FOnBeforeCancel(Self);
end;

procedure TpfgPalmTable.DoAfterCancel;
begin
  if Assigned(FOnAfterCancel) then FOnAfterCancel(Self);
end;

procedure TpfgPalmTable.DoBeforeDelete;
begin
  if Assigned(FOnBeforeDelete) then FOnBeforeDelete(Self);
end;

procedure TpfgPalmTable.DoAfterDelete;
begin
  if Assigned(FOnAfterDelete) then FOnAfterDelete(Self);
end;

procedure TpfgPalmTable.DoBeforeTableCreate;
begin
  if Assigned(FOnBeforeTableCreate) then FOnBeforeTableCreate(Self);
end;

procedure TpfgPalmTable.DoAfterTableCreate;
begin
  if Assigned(FOnAfterTableCreate) then FOnAfterTableCreate(Self);
end;

procedure TpfgPalmTable.DoReadField(pData: Pointer; DataSize: Integer;
  FieldIndex: Integer; var DataPos: LongWord; out Data: string);
begin
  if Assigned(FOnReadField) then
    FOnReadField(Self, pData, DataSize, FieldIndex, DataPos, Data);
end;

procedure TpfgPalmTable.DoWriteField(FieldIndex: Integer; out Data: Variant);
begin
  if Assigned(FOnWriteField) then
    FOnWriteField(Self, FieldIndex, Data);
end;

{==========================================================================}
{= Miscellaneous standard implementations                                 =}
{=                                                                        =}
{==========================================================================}

procedure TpfgPalmTable.Open;
begin
  Active := True;
end;

procedure TpfgPalmTable.Close;
begin
  Active := False;
end;

// FindField
// Returns a reference to the specified field, or nil if not found

function TpfgPalmTable.FindField(FieldName: string): TpfgPalmField;
var
  ctr: Integer;
begin
  for ctr := 0 to FieldCount-1 do
    if AnsiCompareText(Fields[ctr].FieldName, FieldName) = 0 then
    begin
      Result := Fields[ctr];
      Exit;
    end;

  Result := nil;
end;

// FieldByName
// Similar to FindField, but it raises an exception if the specified field
// isn't found

function TpfgPalmTable.FieldByName(FieldName: string): TpfgPalmField;
begin
  Result := FindField(FieldName);
  if not Assigned(Result) then raise EPalmDbError.CreateFmt(SFieldNotFound, [FieldName]);
end;

function TpfgPalmTable.Exists: Boolean;
begin
  Result := False;
end;

{==========================================================================}
{= Property handler methods                                               =}
{=                                                                        =}
{==========================================================================}

procedure TpfgPalmTable.SetFieldDefs(AFieldDefs: TpfgPalmFieldDefs);
begin
  FFieldDefs.Assign(AFieldDefs);
end;

procedure TpfgPalmTable.SetFields(AFields: TpfgPalmFields);
begin
  FFields.Assign(AFields);
end;

function TpfgPalmTable.GetFieldCount: Integer;
begin
  Result := FFieldDefs.Count;
end;

function TpfgPalmTable.GetReadOnly: Boolean;
begin
  Result := FReadOnly;
end;

procedure TpfgPalmTable.SetReadOnly(AReadOnly: Boolean);
begin
  FReadOnly := AReadOnly;
end;

{**************************************************************************}
{* TpfgPalmRemoteTable class                                                 *}
{*                                                                        *}
{**************************************************************************}

constructor TpfgPalmRemoteTable.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FHandle := 0;                        // Inactive
  FAppInfo := TpfgModifiedMemoryStream.Create; // Create stream for app info
  FData := TStringList.Create;              // list for storing record data
  FAppInfo.Modified := False;
  FReadOnly := False;                  // Read only flag
  FTableName := '';                    // Table name
  FVersion := $0100;                   // Table version (1.0)
  FTableType := 'DATA';                // Default table type
  FCategoryFilter := -1;               // No category filtering
  FOptions := [];
  FState := psInactive;                // Table is inactive
  FdbFlags := [];
  FRecordFlags := 0;
  FRecordID := $FFFFFFFF;
  FRecordEmpty := False;
  FCardNum := CARD_NUM;

  FDbSaving := False;

  FRecNum := -1;
  FBOF := True; FEOF := True;          // Flags for at beginning/end of file

  // If the component has just been dropped on a form at design time, then
  // try and grab a copy of the conduit's creator ID
  FCreatorID := '';
  if csDesigning in ComponentState then
    FCreatorID := ConduitCreatorID;
end;

destructor TpfgPalmRemoteTable.Destroy;
begin
  // If the database is active, then close it - any exceptions are
  // suppressed, since we can't abort out of a component destruction
  if Active then
  begin
    try
      CloseDatabase;
    except
    end;
  end;

  FAppInfo.Free;
  FData.Free;

  inherited Destroy;
end;

{==========================================================================}
{= Overriden inherited methods                                            =}
{=                                                                        =}
{==========================================================================}

// Loaded
// After the component is loaded, check if there is a Creator ID assigned,
// and if not, attempt to dynamically assign one

procedure TpfgPalmRemoteTable.Loaded;
begin
  inherited Loaded;

  if FCreatorID = '' then FCreatorID := ConduitCreatorID;
end;

{==========================================================================}
{= Junction methods for embedded methods                                  =}
{=                                                                        =}
{==========================================================================}

// FieldDefChanged
// When the field definitions change, update the fields list

procedure TpfgPalmRemoteTable.FieldDefChanged;
begin
  FFields.Update;
end;

// GetFieldValue
// Gets a field value from a loaded record

function TpfgPalmRemoteTable.GetFieldValue(Index: Integer): Variant;
var
  ctr: Integer;
  s: string;
  dt: TDateTime;
  si: Single;
  d: Double;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);

  case FFieldDefs[Index].DataType of
    ptByte, ptWord, ptLong:
      if FData[Index] = '' then Result := 0
      else Result := StrToInt(FData[Index]);

    ptString, ptInt64:
      // Int64s are treated like strings, since a variant can't hold it
      Result := FData[Index];

    ptSingle:
    begin
      if FData[Index] = '' then si := 0.0
      else si := StrToFloat(FData[Index]);
      Result := si;
    end;

    ptDouble:
    begin
      if FData[Index] = '' then d := 0.0
      else d := StrToFloat(FData[Index]);
      Result := d;
    end;

    ptDate, ptTime, ptDateTime:
    begin
      if FData[Index] = '' then Result := ''
      else
      begin
        dt := StrToDateTime(FData[Index]);
        Result := dt;
      end;
    end;

    ptCustom:
    begin
      s := FData[Index];
      Result := VarArrayCreate([0, Length(s)-1], varByte);
      for ctr := 1 to Length(s) do
        Result[ctr-1] := Byte(s[ctr]);
    end;
  end;
end;

procedure TpfgPalmRemoteTable.SetFieldValue(Index: Integer; Value: Variant);
var
  ctr: Integer;
  Low: Integer;
  s: string;
begin
  if (not DbSaving) and (not Active) then raise EPalmDbError.Create(SNotActiveError);
  if ReadOnly then raise EPalmDbError.Create(SReadOnlyError);

  if not VarIsArray(Value) then
    FData[Index] := Value
  else
  begin
    Low := VarArrayLowBound(Value, 1);
    SetLength(s, VarArrayHighBound(Value, 1) - Low + 1);
    for ctr := 1 to Length(s) do
      s[ctr] := Chr(Byte(Value[ctr+Low-1]));
    FData[Index] := s;
  end;
end;

function TpfgPalmRemoteTable.GetRecordFlags: Byte;
begin
  Result := FRecordFlags;
end;

procedure TpfgPalmRemoteTable.SetRecordFlags(AFlags: Byte);
begin
  FRecordFlags := AFlags;
end;

function TpfgPalmRemoteTable.GetRecordID: LongWord;
begin
  Result := FRecordID;
end;

procedure TpfgPalmRemoteTable.SetRecordID(Value: LongWord);
begin
  FRecordID := Value;
end;

function TpfgPalmRemoteTable.GetRecordCategory: Shortint;
begin
  Result := FRecordCategory;
end;

procedure TpfgPalmRemoteTable.SetRecordCategory(ACategory: Shortint);
begin
  FRecordCategory := ACategory;
end;

function TpfgPalmRemoteTable.GetIsEmpty: Boolean;
begin
  Result := FRecordEmpty;
end;

{==========================================================================}
{= Property handler methods                                               =}
{=                                                                        =}
{==========================================================================}

// SetHandle
// Sets the HotSync Palm database to use

procedure TpfgPalmRemoteTable.SetHandle(AHandle: Byte);
begin
  if FHandle <> AHandle then
  begin
    if FHandle <> 0 then Close;             // Close the active database
    FHandle := AHandle;
  end;
end;

// GetActive
// Returns true if the database is active

function TpfgPalmRemoteTable.GetActive: Boolean;
begin
  Result := FHandle <> 0;
end;

// SetActive
// Sets whether the database is active

procedure TpfgPalmRemoteTable.SetActive(AActive: Boolean);
begin
  if GetActive <> AActive then
  begin
    if csDesigning in ComponentState then raise
      EPalmDbError.Create(SRuntimeError);

    if AActive then
    begin
      DoBeforeOpen;
      OpenDatabase;
      FState := psBrowse;
      DoAfterOpen;
    end
    else
    begin
      DoBeforeClose;
      CloseDatabase;
      FState := psInactive;
      DoAfterClose;
    end;
  end;
end;

function TpfgPalmRemoteTable.GetTableName: string;
begin
  Result := FTableName;
end;

procedure TpfgPalmRemoteTable.SetTableName(AName: string);
begin
  if AName <> FTableName then
  begin
    Close;                        // Make sure database is closed
    FTableName := AName;
  end;
end;

// SetCreatorID
// Sets the creator ID to use

procedure TpfgPalmRemoteTable.SetCreatorID(AID: string);
begin
  if FCreatorID = AID then Exit;
  Close;                          // Make sure database is closed

  // Try and convert the creator ID to a value - this will generate an
  // exception if the creator id is malformed
  StrToCreatorID(AID);

  FCreatorID := AID;
end;

// SetTableType
// Sets the tabletype

procedure TpfgPalmRemoteTable.SetTableType(AType: string);
begin
  if FTableType = AType then Exit;
  Close;                          // Make sure database is closed

  if not IsCreatorID(AType) then
    raise EPalmDbError.CreateFmt(STableTypeError, [AType]);

  FTableType := AType;
end;

procedure TpfgPalmRemoteTable.SetVersion(AVersion: Word);
begin
  if FVersion <> AVersion then
  begin
    Close;                        // Make sure database is closed
    FVersion := AVersion;
  end;
end;

function TpfgPalmRemoteTable.GetOptions: TpfgPalmRemoteTableOptions;
begin
  Result := FOptions;
end;

procedure TpfgPalmRemoteTable.SetOptions(AOptions: TpfgPalmRemoteTableOptions);
begin
  if AOptions <> FOptions then
  begin
    Close;                        // Make sure database is closed
    FOptions := AOptions;
  end;
end;

// GetAppInfo
// Returns the AppInfo block pointer

function TpfgPalmRemoteTable.GetAppInfo: TpfgModifiedMemoryStream;
begin
  Result := FAppInfo;
end;

// SetAppInfo
// Sets the database application info block

procedure TpfgPalmRemoteTable.SetAppInfo(AInfo: TpfgModifiedMemoryStream);
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  if ReadOnly then raise EPalmDbError.Create(SReadOnlyError);

  FAppInfo.Clear;
  FAppInfo.LoadFromStream(AInfo);
end;

// GetCategoryFilter
// Gets the current active cateogry filter (-1 for none)

function TpfgPalmRemoteTable.GetCategoryFilter: Shortint;
begin
  Result := FCategoryFilter;
end;

// SetCategoryFilter
// Sets the active category filter. Valid category numbers are from 0 - 14,
// or -1 for no filtering

procedure TpfgPalmRemoteTable.SetCategoryFilter(ACategory: Shortint);
begin
  if FCategoryFilter = ACategory then Exit;

  if not Active then raise EPalmDbError.Create(SNotActiveError);
  if State in [psEdit, psInsert] then raise EPalmDbError.Create(SEditModeError);

  // Validate that the category is correct
  if not ((ACategory = -1) or ((ACategory >= 0) and (ACategory <= 14))) then
    raise EPalmDbError.CreateFmt(SCategoryError, [ACategory]);

  FCategoryFilter := ACategory;
  DoBeforeRecChange;
  GetNextRecord;
  DoAfterRecChange;
end;

// GetState
// Returns the active state of the table

function TpfgPalmRemoteTable.GetState: TpfgPalmDataSetState;
begin
  Result := FState;
end;

function TpfgPalmRemoteTable.GetRecordCount: Integer;
var
  NumRecs: Word;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);

  PalmCheck(SyncGetDbRecordCount(FHandle, NumRecs));
  Result := NumRecs;
end;

function TpfgPalmRemoteTable.GetRecNum: Integer;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);

  Result := FRecNum;
end;

procedure TpfgPalmRemoteTable.SetRecNum(ARecNum: Integer);
var
  info: CRawRecordInfo;
  rCount: Integer;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  if State in [psEdit, psInsert] then raise EPalmDbError.Create(SEditModeError);

  DoBeforeRecChange;

  rCount := RecordCount;
  if (ARecNum < 0) or (ARecNum > rCount) then
    raise EPalmDbError.CreateFmt(SInvalidRecNumError, [ARecNum]);

  FRecNum := ARecNum;
  if FRecNum = rCount then
  begin
    InitRecord;
    FEOF := True;
  end;

  // Move to the specified record
  GetMem(info.m_pBytes, MaxRecordSize);
  try
    info.m_TotalBytes := MaxRecordSize;
    info.m_FileHandle := FHandle;
    info.m_RecIndex := ARecNum;
    info.m_Attribs := 0;
    info.m_dwReserved := 0;

    PalmCheck(SyncReadRecordByIndex(info));
    FRecordCategory := info.m_catID;
    FRecordID := info.m_recId;

    RemoteRecordToLocal(info.m_pBytes, info.m_RecSize);

    if FRecNum <> 0 then FBOF := False;
    DoAfterRecChange;

  finally
    Dispose(info.m_pBytes);
  end;
end;

// GetBOF
// Returns true if at the beginning of the table

function TpfgPalmRemoteTable.GetBOF: Boolean;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);

  Result := FRecNum = 0;
//  Result := FBOF;
end;

// GetEOF
// Returns true if at the end of the file

function TpfgPalmRemoteTable.GetEOF: Boolean;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);

  Result := FRecNum >= RecordCount;
  // Result := FEOF;
end;

// SetDbFlags
// Sets the database flags for the table

procedure TpfgPalmRemoteTable.SetDbFlags(Value: TpfgPalmRemoteTableFlags);
begin
  // Error checking added because some people didn't realise that the
  // database flags could only be set prior to creating the table
  if Active then raise EPalmDbError.Create(SActiveError);
  FdbFlags := Value;
end;

// SetDbSaving
// Sets the DbSaving flag. If it is being turned on, also calls the table's
// InitRecord method to set up an empty dataset

procedure TpfgPalmRemoteTable.SetDbSaving(Value: Boolean);
begin
  FDbSaving := Value;
  if FDbSaving then InitRecord;
end;

{==========================================================================}
{= Internal support methods                                               =}
{=                                                                        =}
{==========================================================================}

// OpenDatabase
// Opens the remote Palm database

procedure TpfgPalmRemoteTable.OpenDatabase;
var
  openMode: Byte;
  RetVal: Integer;
//  d: CDbCreateDB;
  info: CDbGenInfo;
begin
  { Open up the database for access }
  // Work out the open mode
  openMode := eDbRead;
  if not ReadOnly then openMode := openMode or eDbWrite;
  if (opShowSecret in FOptions) then openMode := openMode or eDbShowSecret;

  // Try and open the table
  RetVal := SyncOpenDb(PChar(FTableName), FCardNum, FHandle, openMode);
  if (RetVal <> 0) {and (RetVal <> SYNCERR_NOT_FOUND)} then PalmError(RetVal);

  // Read in the AppInfo block
  GetMem(info.m_pBytes, MaxRecordSize);
  try
    info.m_TotalBytes := MaxRecordSize;
    info.m_dwReserved := 0;
    SyncReadDbAppInfoBlock(FHandle, info);

    // Store it in the app info stream
    ReallocMem(info.m_pBytes, info.m_BytesRead);
    FAppInfo.Clear;
    FAppInfo.Write(info.m_pBytes^, info.m_BytesRead);
  finally
    Dispose(info.m_pBytes);
  end;

  // Initialize the record set
  InitRecord;

  // Get in the first record, if any
  FRecNum := 0;
  GetNextRecord;
end;

procedure TpfgPalmRemoteTable.CloseDatabase;
var
  RetVal: Integer;
  info: CDbGenInfo;
begin
  // Save app info if modified
  if FAppInfo.Modified then
  begin
    // Fill out parameter block
    info.m_pBytes := FAppInfo.Memory;
    info.m_TotalBytes := FAppInfo.Size;
    info.m_BytesRead := FAppInfo.Size;
    info.m_dwReserved := 0; //Reserved
    SyncWriteDBAppInfoBlock(FHandle, info);
  end;

  // Close the database
  RetVal := SyncCloseDb(FHandle);
  FHandle := 0;
  if RetVal <> 0 then PalmError(RetVal);
end;

// GetNextRecord
// Reads in the next record from the database

procedure TpfgPalmRemoteTable.GetNextRecord;
var
  info: CRawRecordInfo;
  RetVal: Integer;
begin
  GetMem(info.m_pBytes, MaxRecordSize);

  try
    info.m_FileHandle := FHandle;
    info.m_TotalBytes := MaxRecordSize;
    info.m_dwReserved := 0;

    if CategoryFilter <> -1 then
    begin
      info.m_CatId := FCategoryFilter;
      if opModified in FOptions then
        RetVal := SyncReadNextModifiedRecInCategory(info)
      else
        RetVal := SyncReadNextRecInCategory(info);
    end
    else
    begin
      info.m_CatId := 0;

      if opModified in FOptions then
        RetVal := SyncReadNextModifiedRec(info)
      else
      begin
        // Get the specified record
        info.m_RecIndex := FRecNum;
        RetVal := SyncReadRecordByIndex(info);
      end;
    end;

    // Reset the record saved flag
    FRecordSaved := False;

    // Check the result
    if RetVal = SYNCERR_NOT_FOUND then
    begin
      // Must be end of table
      FRecNum := RecordCount;
      FRecordFlags := 0;
      FRecordID := 0;
      FRecordCategory := FCategoryFilter;
      Exit;
    end
    else if RetVal <> 0 then PalmError(RetVal);

    // Retrieve the data from the remote record
    RemoteRecordToLocal(info.m_pBytes, info.m_RecSize);

    FRecordFlags := info.m_Attribs;
    FRecordID := info.m_RecId;
    FRecordCategory := info.m_CatId;

    FBOF := False;
  finally
    Dispose(info.m_pBytes);
  end;
end;

// PostRecord
// Saves the edited record to the database. Checks the current state to
// determine whether it's a new record or an existing one

procedure TpfgPalmRemoteTable.PostRecord;
var
  info: CRawRecordInfo;
  rec: Pointer;
  recSize: LongWord;
begin
  LocalToRemoteRecord(rec, recSize);
  try
    FillChar(info, Sizeof(info), 0);
    info.m_FileHandle := FHandle;
    if FState = psEdit then info.m_RecId := FRecordID  // Existing record
    else info.m_RecID := 0;                            // New one

    info.m_Attribs := FRecordFlags;                 // Record attributes

    if FRecordCategory = -1 then info.m_CatId := 0
    else info.m_CatId := FRecordCategory;
    info.m_pBytes := rec;
    info.m_RecSize := recSize;
    info.m_TotalBytes := recSize;
    PalmCheck(SyncWriteRec(info));
    FRecordID := info.m_RecId;      // Read back the new record ID
    // Read out the record number if it was an insert
    if FState = psInsert then FRecNum := RecordCount;
  finally
    Dispose(rec);
  end;
end;

// DeleteRecord
// Deletes the record with the specified record ID

procedure TpfgPalmRemoteTable.DeleteRecord(ARecID: LongWord);
var
  info: CRawRecordInfo;
begin
  DoBeforeDelete;

  FillChar(info, Sizeof(info), 0);
  info.m_FileHandle := FHandle;
  info.m_RecID := ARecID;
  PalmCheck(SyncDeleteRec(info));

  DoAfterDelete;
end;

// InitRecord
// Initializes the record buffer. This basically involves filling up the
// data string list with a blank entry for each field

procedure TpfgPalmRemoteTable.InitRecord;
var
  ctr: Integer;
begin
  FData.Clear;
  for ctr := 1 to FieldDefs.Count do
    FData.Add('');
end;

// RemoteRecordToLocal
// This method takes a remote record read in from the Palm, and using the
// current field definitions, converts the data it contains into a list of
// strings stored in the FData string list

procedure TpfgPalmRemoteTable.RemoteRecordToLocal(const pData: Pointer; Size: LongWord);
var
  Index: Integer;
  ptr: LongWord;
  FieldSize: Integer;
  p: PByte;
  sValue: string;
  Data: string;
  vSingle: LongWord;
  vDouble: Int64;
  pLW: PLongWord;
  v: Int64;
begin
  InitRecord;

  // Check whether the record is empty
  FRecordEmpty := Size = 0;
  if FRecordEmpty then Exit;

  ptr := 0;
  for Index := 0 to FieldDefs.Count-1 do
  begin
    // Check to see whether a field handler has been set
    if Assigned(FOnReadField) then
    begin
      try
        FOnReadField(Self, pData, Size, Index, ptr, Data);
        FData[Index] := Data;

        // Field handled, so move onto next field
        Continue;

      except
        // Suppress any abort and fall back on standard field handling
        on EAbort do
        begin
        end;
      end;
    end;

    // If number aligning is on, check if the field is a 2 or 4 byte number,
    // and if so, make sure it's on a word boundary
    if (opNumAlign in FOptions) and (ptr mod 2 = 1) and
       (PalmFieldSize[FieldDefs[Index].DataType] mod 2 = 0) then
      Inc(ptr);

    // Work out the field size
    FieldSize := PalmFieldSize[FieldDefs[Index].DataType];
    if FieldSize = -1 then FieldSize := FieldDefs[Index].Size;

    // Make sure the field is within the physical record
    if ptr >= LongWord(Size) then
    begin
      if not FieldDefs.StrictFields then Break   // Finished mapping fields
      else raise EPalmDbError.Create(SFieldMappingError);
    end;
    if (FieldSize > 0) and (ptr + LongWord(FieldSize) > LongWord(Size)) then
      raise EPalmDbError.Create(SFieldMappingError);

    // Get the value
    p := PByte(LongWord(pData) + ptr);

    if FieldDefs[Index].DataType = ptString then
    begin
      { String type }
      if (FieldSize = 0) or ((FieldSize > 0) and
         (Integer(StrLen(PChar(p))) < FieldSize)) then
      begin
        sValue := StrPas(PChar(p))
      end
      else
      begin
        // Special case: fixed size string field completely full up
        SetLength(sValue, FieldSize);
        Move(PChar(p)^, sValue[1], FieldSize);
      end;

      // Increment the pointer
      if FieldSize > 0 then Inc(ptr, FieldSize)
      else Inc(ptr, Length(sValue)+1); // Variable length string, past NULL
    end
    else
    begin
      { Handle remaining data types }
      case FieldDefs[Index].DataType of
        ptByte: sValue := IntToStr(p^);
        ptWord: sValue := IntToStr(ReverseVal(PWord(p)^, 2));
        ptLong: sValue := IntToStr(Integer(ReverseVal(PInteger(p)^, 4)));
        ptDate: sValue := DateToStr(PalmDateToDateTime(ReverseVal(PWord(p)^, 2)));
        ptTime: sValue := PalmTimeToTimeStr(PTimeType(p)^);
        ptDateTime: sValue := DateTimeToStr(PalmDateTimeToDateTime(pDateTimeType(p)^));
        ptInt64:
        begin
          // The conversion is somewhat broken up because the combined
          // version kept crashing Kernel32.dll for some reason
          pLW := PLongWord(p);
          v := ReverseVal(pLW^, 4) shl 32;
          Inc(pLW);
          v := v + ReverseVal(pLW^, 4);

          sValue := IntToStr(v);
        end;
        ptSingle:
        begin
          vSingle := ReverseVal(PLongWord(p)^, 4);
          sValue := FloatToStr(PSingle(@vSingle)^);
        end;
        ptDouble:
        begin
          vDouble := ReverseVal(PInt64(p)^, 8);
          sValue := FloatToStr(PDouble(@vDouble)^);
        end;
      else
        // ptCustom: Unknown data block, so store as a representative string
        SetLength(sValue, FieldSize);
        Move(PChar(p)^, sValue[1], FieldSize);
      end;

      Inc(ptr, FieldSize);
    end;

    // Add the string value into the data list
    FData[Index] := sValue;

    // Handle string alignment processisng
    if FieldDefs[Index].DataType = ptString then
    begin
      if (not (opStringsAlign in Options)) or
         ((Index < FieldDefs.Count-1) and (FieldDefs[Index+1].DataType = ptString)) then
        Continue;
    end;
    
    // Make sure the pointer is on an alignment boundary for the next field,
    // if there is an alignment entry setting
    if (FieldDefs.Alignment <> 0) and
       (ptr mod FieldDefs.Alignment <> 0) then
      ptr := ((ptr + FieldDefs.Alignment - 1) div FieldDefs.Alignment) *
        FieldDefs.Alignment;
  end;
end;

// LocalToRemoteRecord
// This method creates a data block that represents the Palm record for the
// current field data contained in the FData string list. It is the
// responsibility of the caller to free the memory when finished with it

procedure TpfgPalmRemoteTable.LocalToRemoteRecord(out pData: Pointer; out Size: LongWord);
const
  SRecordError = 'LocalToRemoteRecord: Record went beyond it''s limits';
var
  Index, ctr: Integer;
  FieldSize: Integer;
  ptr: LongWord;
  p: PByte;
  fsize: Integer;
  vData: Variant;
  vSingle: Single;
  vDouble: Double;
  v: Int64;
  pLW: PLongWord;

  function StrToInt(const s: string): Integer;
  begin
    if s = '' then Result := 0
    else Result := SysUtils.StrToInt(s);
  end;

  function StrToInt64(const s: string): Int64;
  begin
    if s = '' then Result := 0
    else Result := SysUtils.StrToInt64(s);
  end;

  function StrToFloat(const s: string): Double;
  begin
    if s = '' then Result := 0.0
    else Result := SysUtils.StrToFloat(s);
  end;

  function StrToDate(const S: string): TDateTime;
  begin
    if s = '' then Result := 0
    else Result := SysUtils.StrToDate(s);
  end;

  function StrToDateTime(const S: string): TDateTime;
  begin
    if s = '' then Result := 0
    else Result := SysUtils.StrToDateTime(s);
  end;

begin
  // Firstly we need to work out how much memory needs to be allocated, by
  // totalling up the size of each individual field
  Size := 0;
  for Index := 0 to FieldDefs.Count-1 do
  begin
    // If a write field event handler is set, use it in preference
    if Assigned(FOnWriteField) then
    begin
      // At this stage we are merely finding out how much space to allocate
      // for storage; we will later loop again to get the actual data. This
      // means that the data returned from looping through the fields
      // multiple times must remain identical
      try
        FonWriteField(Self, Index, vData);
        Inc(Size, VarArrayHighBound(vData, 1)+1);
        Continue;

      except
        // Suppress any aborts and allow standard field handling to go forward
        on EAbort do
        begin
        end;
      end;
    end;

    // Get the field size
    FieldSize := PalmFieldSize[FieldDefs[Index].DataType];
    if FieldSize = -1 then FieldSize := FieldDefs[Index].Size;

    if FieldSize > 0 then Inc(Size, FieldSize)
    else
    begin
      // Variable length string, or custom data block. For strings, the size
      // is the string length plus 1 for the NULL terminator. For custom data
      // blocks, it's just the length of the data itself
      if FieldDefs[Index].DataType = ptString then
        Inc(Size, Length(FData[Index])+1)
      else
        Inc(Size, Length(FData[Index]));
    end;

    // Alignment control handler for variable length strings
    if (FieldDefs[Index].DataType = ptString) and
       (FieldDefs[Index].Size = 0) then
    begin
      // Don't perform alignment control if the following field is also
      // a string, or if the string alignment flag is not set in the options
      if ((Index+1 < FieldDefs.Count) and (FieldDefs[Index+1].DataType = ptString)) or
         (not (opStringsAlign in Options)) then
        // Skip over alignment adjust
        Continue;
    end;

    // Handle rounding up to alignment boundary
    if (FieldDefs.Alignment <> 0) and
       (Size mod FieldDefs.Alignment <> 0) then
      Size := ((Size + FieldDefs.Alignment - 1) div FieldDefs.Alignment) *
        FieldDefs.Alignment;
    if (opNumAlign in FOptions) and (Size mod 2 <> 0) and
       (PalmFieldSize[FieldDefs[Index].DataType] mod 2 = 0) then
      Inc(Size);
  end;

  // Alloacte the memory for the record
  GetMem(pData, Size);

  // Loop through filling out the contents of the record with each field
  ptr := 0;
  for Index := 0 to FieldDefs.Count-1 do
  begin
    // Get a pointer to the start of field
    p := PByte(LongWord(pData) + ptr);

    // If a write field event handler is set, use it to get the data
    if Assigned(FOnWriteField) then
    begin
      try
        FOnWriteField(Self, Index, vData);
        for ctr := 0 to VarArrayHighBound(vData, 1) do
        begin
          if ptr >= Size then raise Exception.Create(SRecordError);
          p^ := vData[ctr];
          Inc(p); Inc(ptr);
        end;
        Continue;

      except
        on EAbort do
        begin
        end;
      end;

    end;

    // Get the field size
    FieldSize := PalmFieldSize[FieldDefs[Index].DataType];
    if FieldSize = -1 then FieldSize := FieldDefs[Index].Size;

    // Handle word alignment if dealing with a word or long value
    if (opNumAlign in FOptions) and (ptr mod 2 <> 0) and
       (PalmFieldSize[FieldDefs[Index].DataType] mod 2 = 0) then
    begin
      Inc(ptr);
      Inc(p);
    end;

    if FieldDefs[Index].DataType = ptString then
    begin
      { String type }
      if FieldSize <= 0 then
      begin
        if ptr + LongWord(Length(FData[Index])) > Size then
          Exception.Create(SRecordError);
        StrPCopy(PChar(p), FData[Index]);
        Inc(ptr, Length(FData[Index])+1);
      end
      else
      begin
        if ptr + LongWord(FieldSize) > Size then Exception.Create(SRecordError);
        FillChar(p^, FieldSize, 0); // Blank out the entire strings
        StrLCopy(PChar(p), PChar(FData[Index]), FieldSize);
        Inc(ptr, FieldSize);
      end;
    end
    else
    begin
      { Handle remaining data types }
      if ptr + LongWord(FieldSize) > Size then Exception.Create(SRecordError);

      case FieldDefs[Index].DataType of
        ptByte: p^ := StrToInt(FData[Index]);
        ptWord: PWord(p)^ := ReverseVal(StrToInt(FData[Index]), 2);
        ptLong: PDWord(p)^ := ReverseVal(StrToInt(FData[Index]), 4);
        ptDate: PWord(p)^ := ReverseVal(DateTimeToPalmDate(StrToDate(FData[Index])), 2);
        ptTime: PTimeType(p)^ := TimeStrToPalmTime(FData[Index]);
        ptDateTime: DateTimeToPalmDateTime(StrToDateTime(FData[Index]),
                       PDateTimeType(p)^);
        ptInt64:
        begin
          // The code is split up because I suffered from crashes when
          // I tried to directly use a pointer to an Int64
          v := StrToInt64(FData[Index]);
          pLW := PLongWord(p);
          pLW^ := ReverseVal(v shr 32, 4);
          Inc(pLW);
          pLW^ := ReverseVal(v and $ffffffff, 4);
        end;
        ptSingle:
        begin
          vSingle := StrToFloat(FData[Index]);
          PLongWord(p)^ := ReverseVal(PLongWord(@vSingle)^, 4);
        end;
        ptDouble:
        begin
          vDouble := StrToFloat(FData[Index]);
          PInt64(p)^ := ReverseVal(PInt64(@vDouble)^, 8);
        end;
      else
        // ptCustom: Unknown data block, store the exact string data
        fsize := Length(FData[Index]);
        if fsize < FieldDefs[Index].Size then
          FillChar(p^, FieldDefs[Index].Size, 0)
        else if fsize > FieldDefs[Index].Size then fsize := FieldDefs[Index].Size;
        Move(PChar(FData[Index])^, p^, fsize);
      end;

      Inc(ptr, FieldSize);
    end;

    // Alignment control handler for variable length strings
    if (FieldDefs[Index].DataType = ptString) and
       (FieldDefs[Index].Size = 0) then
    begin
      // Don't perform alignment control if the following field is also
      // a string, or if the string alignment flag is not set in the options
      if ((Index+1 < FieldDefs.Count) and (FieldDefs[Index+1].DataType = ptString)) or
         (not (opStringsAlign in Options)) then
        // Skip over alignment adjust
        Continue;
    end;

    // Make sure the pointer is on an alignment boundary for the next field,
    // if there is an alignment entry setting
    if (FieldDefs.Alignment <> 0) and
       (ptr mod FieldDefs.Alignment <> 0) then
      ptr := ((ptr + FieldDefs.Alignment - 1) div FieldDefs.Alignment) *
        FieldDefs.Alignment;
  end;
end;

function TpfgPalmRemoteTable.GetDbFlagsField: Byte;
begin
  Result := eRecord;
  if tfAppInfoDirty in FDBFlags then
     Result := Result + eAppInfoDirty;
  if tfBackupDB in FDBFlags then
     Result := Result + eBackupDB;
  if tfOkToInstallNewer in FDBFlags then
     Result := Result + eOkToInstallNewer;
  if tfCopyPrevention in FDBFlags then
     Result := Result + eCopyPrevention;
  if tfResetAfterInstall in FDBFlags then
     Result := Result + eResetAfterInstall;
end;

{==========================================================================}
{= Standard DB Methods                                                    =}
{=                                                                        =}
{==========================================================================}

// EmptyTable
// Empties the table of all it's contents. Does not require the table to
// be open

procedure TpfgPalmRemoteTable.EmptyTable;
var
  TempHandle: Byte;
  openMode: Byte;
begin
  // Get a handle to the table, opening table if necessary
  if Active then
  begin
    if State in [psEdit, psInsert] then raise EPalmDbError.Create(SEditModeError);
    TempHandle := FHandle
  end
  else
  begin
    openMode := eDbRead or eDbWrite;
    PalmCheck(SyncOpenDb(PChar(FTableName), FCardNum, TempHandle, openMode));
  end;

  // Empty the table
  PalmCheck(SyncPurgeAllRecs(TempHandle));

  if Active then
  begin
    // Reset table access fields
    FRecNum := 0;
    FRecordFlags := 0;
    FRecordID := 0;
    InitRecord;
  end
  else
  begin
    // Close the table
    SyncCloseDb(TempHandle);
  end;
end;

procedure TpfgPalmRemoteTable.CreateTable;
var
  stats: CDbCreateDb;
begin
  if Active then raise EPalmDbError.Create(SActiveError);

  DoBeforeTableCreate;                 // Signal table being created

  FillChar(stats, Sizeof(stats), 0);
  stats.m_Creator := StrToCreatorID(FCreatorID);
  StrLCopy(stats.m_Name, PChar(FTableName), SYNC_DB_NAMELEN-1);

  // Work out the db flags to use
  stats.m_Flags := GetDbFlagsField;

  stats.m_Type := StrToCreatorID(FTableType);
  stats.m_CardNo := FCardNum;
  stats.m_Version := FVersion;
  stats.m_dwReserved := 0;

  PalmCheck(SyncCreateDb(stats));
  FHandle := stats.m_FileHandle;

  DoAfterTableCreate;                 // Signal table has been created

  // Initialize flags
  FState := psBrowse;
  FRecordFlags := 0;
  FRecordID := 0;

  // Initialize record set
  InitRecord;
end;

procedure TpfgPalmRemoteTable.DeleteTable;
begin
  if Active then raise EPalmDbError.Create(SActiveError);

  PalmCheck(SyncDeleteDb(PChar(TableName), FCardNum));
end;

procedure TpfgPalmRemoteTable.First;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  if State in [psEdit, psInsert] then raise EPalmDbError.Create(SEditModeError);

  DoBeforeRecChange;

  PalmCheck(SyncResetRecordIndex(FHandle));
  FRecNum := 0;
  GetNextRecord;

  DoAfterRecChange;
end;

// Last
// Moves to the end of the table. We simulate this by clearing the record

procedure TpfgPalmRemoteTable.Last;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  if State in [psEdit, psInsert] then raise EPalmDbError.Create(SEditModeError);

  FRecNum := RecordCount;
  InitRecord;
end;

// Next
// Moves to the next record

procedure TpfgPalmRemoteTable.Next;
begin
  DoBeforeRecChange;

  // Special check: If a record was posted, and we're not using any
  // filtering, keep the record number the same, so it will correctly
  // get in the record that was following the saved one
  if ((not FRecordSaved) or (CategoryFilter <> -1)) then
    FRecNum := FRecNum + 1;

  GetNextRecord;
  DoAfterRecChange;
end;

// Prior
// Moves to the prior record

procedure TpfgPalmRemoteTable.Prior;
begin
  raise ENotImplError.Create('Prior');
end;

procedure TpfgPalmRemoteTable.Edit;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  if State in [psEdit, psInsert] then raise EPalmDbError.Create(SEditModeError);

  DoBeforeEdit;
  FState := psEdit;
  DoAfterEdit;
end;

procedure TpfgPalmRemoteTable.Append;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  if State in [psEdit, psInsert] then raise EPalmDbError.Create(SEditModeError);

  DoBeforeInsert;
  FState := psInsert;
  FRecordFlags := 0;          // No attributes
  InitRecord;                 // Blank field set
  DoAfterInsert;
end;

procedure TpfgPalmRemoteTable.Insert;
begin
  // Palm provides no method for inserting new records in the middle of a
  // table during synchronization, so fall back on using the Append method
  Append;
end;

procedure TpfgPalmRemoteTable.Post;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  if not (State in [psEdit, psInsert]) then raise EPalmDbError.Create(SEditModeError);

  DoBeforePost;
  PostRecord;
  FRecordSaved := True;          // Flag the record has been saved
  FState := psBrowse;            // Move back into browsing
  DoAfterPost;
end;

procedure TpfgPalmRemoteTable.Cancel;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  if not (State in [psEdit, psInsert]) then raise EPalmDbError.Create(SEditModeError);
  DoBeforeCancel;
  FState := psBrowse;
  DoAfterCancel;
end;

procedure TpfgPalmRemoteTable.Delete;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  if State in [psEdit, psInsert] then raise EPalmDbError.Create(SEditModeError);
  if (FRecordID = 0) or (FRecordID = $FFFFFFFF) then
    raise EPalmDbError.Create(SRecNotFound);

  DeleteRecord(FRecordID);

  // Get the next record from the database
  GetNextRecord;
end;

// DeleteById
// Deletes the record with the specified Record ID from the table

procedure TpfgPalmRemoteTable.DeleteById(ARecID: LongWord);
begin
  // If trying to delete the current record, use the standard Delete method,
  // since it will take care of retrieivng the next record after deleting
  // the current one
  if ARecID = FRecordID then
    Delete
  else
  begin
    if not Active then raise EPalmDbError.Create(SNotActiveError);
    if State in [psEdit, psInsert] then raise EPalmDbError.Create(SEditModeError);

    DeleteRecord(ARecID);
  end;
end;

// LocateById
// Locates the record with the specified record ID from the table. Returns
// true if the record was found 

function TpfgPalmRemoteTable.LocateById(ARecID: LongWord): Boolean;
var
  info: CRawRecordInfo;
  RetVal: integer;
begin
  Result := True;                      // Presume it will be successfull
  if ARecID <> FRecordID then
  begin
    if not Active then raise EPalmDbError.Create(SNotActiveError);
    if State in [psEdit, psInsert] then raise EPalmDbError.Create(SEditModeError);

    DoBeforeRecChange;

    // Move to the specified record
    GetMem(info.m_pBytes, MaxRecordSize);
    try
      info.m_TotalBytes := MaxRecordSize;
      info.m_FileHandle := FHandle;
      info.m_RecId := ARecID;

      info.m_RecIndex := 0;
      info.m_Attribs := 0;
      info.m_dwReserved := 0;

      RetVal := SyncReadRecordByID(info);

      // Check the result
      if RetVal = SYNCERR_NOT_FOUND then
      begin
        // Record could not be found
        FRecNum := 0;
        FRecordFlags := 0;
        FRecordID := 0;
        Result := False;
        Exit;
      end
      else if RetVal <> 0 then PalmError(RetVal);

      // Found the record, so load it in
      FRecordCategory := info.m_catID;
      FRecNum := info.m_RecIndex;
      FRecordID := info.m_RecId;
      FRecordFlags := info.m_Attribs;

      RemoteRecordToLocal(info.m_pBytes, info.m_RecSize);
      DoAfterRecChange;

    finally
      Dispose(info.m_pBytes);
    end;
  end;
end;

// Exists
// Returns true if the table exists

function TpfgPalmRemoteTable.Exists: Boolean;
var
  TempHandle: Byte;
  openMode: Byte;
  RetVal: Integer;
begin
  if Active then
    Result := True
  else
  begin
    openMode := eDbRead;
    RetVal := SyncOpenDb(PChar(FTableName), FCardNum, TempHandle, openMode);

    Result := RetVal = S_OK;
    if Result then SyncCloseDb(TempHandle)
    else if RetVal <> SYNCERR_NOT_FOUND then
      // Some general error occurred, so raise the appropriate error
      PalmCheck(RetVal);
  end;
end;

// PurgeTable
// Purges the table of all the records marked as deleted

procedure TpfgPalmRemoteTable.PurgeTable;
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  SyncPurgeDeletedRecs(FHandle);
end;

// SaveToFile
// Saves the database to the specified local file in a .pdb file format

procedure TpfgPalmRemoteTable.SaveToFile(Filename: string);
begin
  if not Active then raise EPalmDbError.Create(SNotActiveError);
  ExportDatabaseToFile(FHandle, Filename);
end;

// PublicLocalToRemote
// This method simply passes on to the private LocalToRemoteRecord method. It
// is present to allow the DataSet to Palm table image routine in pfgDbSave.pas
// to make use of the table's record packing facilities

procedure TpfgPalmRemoteTable.PublicLocalToRemoteRecord(out pData: Pointer;
  out Size: LongWord);
begin
  LocalToRemoteRecord(pData, Size);
end;

{**************************************************************************}
{* TpfgPalmBatchMove class                                                   *}
{*                                                                        *}
{* This simple class implements a basic batch moving functionality, for   *}
{* moving all the records in a source to the destination                  *}
{**************************************************************************}

constructor TpfgPalmBatchMove.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSource := nil;
  FDestination := nil;
end;

procedure TpfgPalmBatchMove.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if FSource = AComponent then FSource := nil;
    if FDestination = AComponent then FDestination := nil;
  end;
end; 

procedure TpfgPalmBatchMove.Execute;
var
  SrcActive, DestActive: Boolean;
begin
  if (not Assigned(FSource)) or (not Assigned(FDestination)) then
    raise EPalmDbError.Create(SSrcDestMissingError);

  SrcActive := Source.Active;
  DestActive := Destination.Active;
  Source.Open;
  Destination.Open;

  while not Source.Eof do
  begin
    Destination.Append;
    Destination.Fields.Assign(Source.Fields);
    Destination.Post;
    Source.Next;
  end;

  if not SrcActive then Source.Close;
  if not DestActive then Destination.Close;
end;

end.
