unit pfgPollingThread;
{**************************************************************************}
{* pfgPollingThread unit                                                  *}
{*                                                                        *}
{* This unit contains the TpfgPalmPollingThread class, a thread class     *}
{* designed to keep the Palm connection active when the main thread isn't *}
{* able to do so.                                                         *}
{*                                                                        *}
{* Copyright  2001 by Zero.Com Pty Ltd All Rights Reserved               *}
{**************************************************************************}

interface

uses Classes, Windows, SyncObjs;

type
  TpfgPalmPollingThread = class(TThread)
  private
    FCriticalSection: TRTLCriticalSection;
    FEventTerminate: TEvent;
    FInCritical: Boolean;
    FDone: Boolean;

    procedure DummyCall;
  public
    constructor Create; reintroduce;
    destructor Destroy; override;
    procedure Execute; override;

    procedure Terminate;
    procedure Lock;
    procedure Unlock;
    property InCritical: Boolean read FInCritical;
    property Done: Boolean read FDone;
  end;

implementation

uses pfgSyncMgr;

constructor TpfgPalmPollingThread.Create;
begin
  inherited Create(True);
  FDone := False;
  InitializeCriticalSection(FCriticalSection);
  FEventTerminate := TEvent.Create(nil, False, False, '');
  FreeOnTerminate := False;
  FInCritical := False;
end;

destructor TpfgPalmPollingThread.Destroy;
begin
  DeleteCriticalSection(FCriticalSection);
  FEventTerminate.Free;
  inherited Destroy;
end;

procedure TpfgPalmPollingThread.Execute;
const
  WaitPeriod = 15000;
var
  TerminateHandle: THandle;
begin
  TerminateHandle := FEventTerminate.Handle;

  try
    while not Terminated do
    begin
      case WaitForMultipleObjects(1, @TerminateHandle, False, WaitPeriod) of
        WAIT_OBJECT_0:      // Terminate
        begin
        end;

        WAIT_TIMEOUT:
        begin
          // Call a Sync API function to keep the connection active
          EnterCriticalSection(FCriticalSection);
          try
            DummyCall;
          finally
            LeaveCriticalSection(FCriticalSection);
          end;
        end;
      end;
    end;
  except
    // Suppress any errors that occur, since the thread can't deal with them
  end;

  FDone := True;
end;

procedure TpfgPalmPollingThread.Terminate;
begin
  inherited Terminate;
  if FInCritical then
    Unlock;
  FEventTerminate.SetEvent;
end;

procedure TpfgPalmPollingThread.Lock;
begin
  if FInCritical then Exit;
  EnterCriticalSection(FCriticalSection);
  FInCritical := True;
end;

procedure TpfgPalmPollingThread.Unlock;
begin
  if not FInCritical then Exit;
  LeaveCriticalSection(FCriticalSection);
  FInCritical := False;
end;

// DummyCall
// Makes a repeated dummy call to the Palm to keep the connection active.
// Note that SyncYieldCycles isn't used because it doesn't reliably work

procedure TpfgPalmPollingThread.DummyCall;
var
  params: SyncReadOpenDbInfoParams;
  info: SyncDatabaseInfoType;
begin
  params.bOptFlags := 0;
  params.bDbHandle := 0;
  SyncReadOpenDbInfo(params, info);
end;

end.
